from pathlib import Path
from typing import Any, Type
from uuid import UUID

import pytest

from tests.controller.sample_graphdata import (
    maps_in_series,
    simple_eagerifelse,
    simple_eval,
    simple_ifelse,
    simple_loop,
    simple_map,
)
from tests.controller.loop_graphdata import loop_multiple_acc, loop_multiple_acc_untyped
from tests.controller.typed_graphdata import (
    tuple_untuple,
    typed_destructuring,
    typed_eval,
    typed_loop,
    typed_map,
    factorial,
    gcd,
)
from tierkreis.controller import run_graph
from tierkreis.controller.data.location import Loc
from tierkreis.controller.data.types import PType, ptype_from_bytes
from tierkreis.controller.executor.in_memory_executor import InMemoryExecutor
from tierkreis.controller.executor.shell_executor import ShellExecutor
from tierkreis.controller.storage.filestorage import ControllerFileStorage
from tierkreis.controller.storage.in_memory import ControllerInMemoryStorage
from tierkreis.controller.data.graph import GraphData

params: list[tuple[GraphData, Any, str, int, dict[str, Any]]] = [
    (simple_eval(), 12, "simple_eval", 1, {}),
    (simple_loop(), 10, "simple_loop", 2, {}),
    (simple_map(), list(range(6, 47, 2)), "simple_map", 3, {}),
    (maps_in_series(), list(range(0, 81, 4)), "maps_in_series", 4, {}),
    (simple_ifelse(), 1, "simple_ifelse", 6, {"pred": b"true"}),
    (simple_ifelse(), 2, "simple_ifelse", 7, {"pred": b"false"}),
    (factorial().get_data(), 24, "factorial", 8, {"value": b"4"}),
    (
        loop_multiple_acc_untyped(),
        {"acc1": 6, "acc2": 12, "acc3": 18},
        "multi_acc",
        9,
        {},
    ),
    (
        loop_multiple_acc().get_data(),
        {"acc1": 6, "acc2": 12, "acc3": 18},
        "multi_acc",
        9,
        {},
    ),
    (simple_eagerifelse(), 1, "simple_eagerifelse", 10, {"pred": b"true"}),
    (factorial().get_data(), 120, "factorial", 12, {"value": b"5"}),
    (typed_eval().get_data(), 12, "typed_eval", 14, {}),
    (typed_loop().get_data(), 10, "typed_loop", 15, {}),
    (typed_map().get_data(), list(range(6, 47, 2)), "typed_map", 16, {}),
    (typed_destructuring().get_data(), list(range(6, 47, 2)), "typed_map", 17, {}),
    (tuple_untuple().get_data(), 3, "tuple_untuple", 18, {}),
    (gcd().get_data(), 21, "gcd", 19, {"a": 1071, "b": 462}),
    (gcd().get_data(), 2, "gcd", 20, {"a": 12, "b": 26}),
    (gcd().get_data(), 24, "gcd", 21, {"a": 48, "b": 360}),
    (gcd().get_data(), 1, "gcd", 22, {"a": 9357, "b": 5864}),
    (gcd().get_data(), 3, "gcd", 23, {"a": 3, "b": 0}),
]
ids = [
    "simple_eval",
    "simple_loop",
    "simple_map",
    "maps_in_series",
    "simple_ifelse_true",
    "simple_ifelse_false",
    "factorial_4",
    "loop_multiple_acc_untyped",
    "loop_multiple_acc",
    "simple_eagerifelse",
    "factorial_5",
    "typed_eval",
    "typed_loop",
    "typed_map",
    "typed_destructuring",
    "tuple_untuple",
    "gcd_1071_462",
    "gcd_12_26",
    "gcd_48_360",
    "gcd_9357_5864",
    "gcd_3_0",
]

storage_classes = [ControllerFileStorage, ControllerInMemoryStorage]
storage_ids = ["FileStorage", "In-memory"]


@pytest.mark.parametrize("storage_class", storage_classes, ids=storage_ids)
@pytest.mark.parametrize("graph,output,name,id,inputs", params, ids=ids)
def test_resume_eval(
    storage_class: Type[ControllerFileStorage | ControllerInMemoryStorage],
    graph: GraphData,
    output: Any,
    name: str,
    id: int,
    inputs: dict[str, PType],
):
    g = graph
    storage = storage_class(UUID(int=id), name=name)
    executor = ShellExecutor(Path("./python/examples/launchers"), Path(""))
    if isinstance(storage, ControllerInMemoryStorage):
        executor = InMemoryExecutor(Path("./tierkreis/tierkreis"), storage=storage)
    storage.clean_graph_files()
    run_graph(storage, executor, g, inputs)

    output_ports = g.nodes[g.output_idx()].inputs.keys()
    actual_output = {}
    for port in output_ports:
        actual_output[port] = ptype_from_bytes(storage.read_output(Loc(), port))

    if f"{name}_output" in actual_output:
        assert actual_output[f"{name}_output"] == output
    elif "value" in actual_output:
        assert actual_output["value"] == output
    else:
        assert actual_output == output



from tierkreis.labels import Labels
from tierkreis.worker.worker import Worker
from tierkreis.value import Value
from tierkreis.controller import run_graph

__all__ = ["Labels", "Worker", "Value", "run_graph"]



from pathlib import Path

# from typing import Callable
# from tierkreis.controller.executor.hpc.hpc_executor import run_hpc_executor
from tierkreis.controller.executor.hpc.job_spec import JobSpec


_COMMAND_PREFIX = "#PBS"


def generate_pbs_script(spec: JobSpec) -> str:
    # 1. Shebang and file header
    lines = [
        """#!/bin/bash
#
# PBS Job Script generated by TIERKREIS

# --- Core Job Specifications ---"""
    ]
    # 2. Name
    lines.append(f"{_COMMAND_PREFIX} -N {spec.job_name}")
    # 3. Resources (node exclusive)
    lines.append(f"{_COMMAND_PREFIX} -l walltime={spec.walltime}")
    resources = f"{_COMMAND_PREFIX} -l select={spec.resource.nodes}"
    if spec.resource.cores_per_node is not None:
        resources += f":ncpus={spec.resource.cores_per_node}"
    if spec.resource.memory_gb is not None:
        resources += f":mem={spec.resource.memory_gb}gb"
    if spec.resource.gpus_per_node is not None:
        resources += f":ngpus={spec.resource.gpus_per_node}"
    if spec.mpi is not None and spec.mpi.max_proc_per_node is not None:
        resources += f":mpiprocs={spec.mpi.max_proc_per_node}"
    if spec.container is not None:
        resources += f":container_engine={spec.container.engine}"
    lines.append(resources)
    if spec.queue is not None:
        lines.append(f"{_COMMAND_PREFIX} -q {spec.queue}")

    # 4. User settings
    lines.append("\n# --- User Details ---")
    if spec.account is not None:
        lines.append(f"{_COMMAND_PREFIX} -A {spec.account}")
    if spec.user is not None:
        if spec.user.mail is not None:
            lines.append(f"{_COMMAND_PREFIX} -m e")  # end only
            lines.append(f"{_COMMAND_PREFIX} -M {spec.user.mail}")

    # 5. Output and Error handling
    lines.append("\n# --- Output and Error Handling ---")
    if spec.error_path is not None:
        lines.append(f"{_COMMAND_PREFIX} -e {spec.error_path}")
    else:
        lines.append(f"{_COMMAND_PREFIX} -j oe")
    if spec.output_path is None:
        spec.output_path = Path(f"./{spec.job_name}.o%j")
    lines.append(f"{_COMMAND_PREFIX} -o {spec.output_path}")

    # 6. MPI
    if spec.mpi is not None:
        lines.append("\n# --- MPI ---")
        lines.append("\n MPI is covered in -l resource declaration")

    # 7. User specific
    lines.append("\n# --- User-Specific Arguments ---")
    for key, value in spec.extra_scheduler_args.items():
        lines.append(f"{_COMMAND_PREFIX} {key} {value if value is not None else ''}")

    # 8. Environment
    lines.append("\n# --- Environment Setup ---")
    if spec.environment != {}:
        env = ",".join(
            f"{key}={value if value else '""'}"
            for key, value in spec.environment.items()
        )
        lines.append(f"-v w{env}")
    # 9. Container logic # taken from nscc docs for enroot:
    if spec.container is not None:
        lines.append(f"{_COMMAND_PREFIX} -l container_image={spec.container.image}")
        if spec.container.name is not None:
            lines.append(f"{_COMMAND_PREFIX} -l container_name={spec.container.name}")
        for key, value in spec.container.extra_args.items():
            lines.append(f"{_COMMAND_PREFIX} -l {key}={value}")
        if spec.container.env_file is not None:
            lines.append(
                f"{_COMMAND_PREFIX} -l {spec.container.engine}_env_file={spec.container.env_file}"
            )  # check if this makes sense for others beside enroot

    # 10. User Command, (prologue), command, (epilogue)
    lines.append("\n# --- User Command ---")
    lines.append(spec.command)

    return "\n".join(lines)


# Disabled for now, needs testing with a PBS system, will be re-enabled later
# See: Issue #182
# class PBSExecutor:
#     def __init__(
#         self,
#         registry_path: Path | None,
#         logs_path: Path,
#         spec: JobSpec,
#         command: str = "qsub",
#     ) -> None:
#         self.launchers_path = registry_path
#         self.logs_path = logs_path
#         self.errors_path = logs_path
#         self.spec = spec
#         self.script_fn: Callable[[JobSpec], str] = generate_pbs_script
#         self.command = command

#     def run(self, launcher_name: str, worker_call_args_path: Path) -> None:
#         self.errors_path = (
#             self.logs_path.parent.parent / worker_call_args_path.parent / "errors"
#         )
#         run_hpc_executor(self, launcher_name, worker_call_args_path)



from pathlib import Path
from uuid import UUID

from tierkreis.builder import GraphBuilder
from tierkreis.controller import run_graph
from tierkreis.controller.data.core import EmptyModel
from tierkreis.controller.data.location import Loc
from tierkreis.controller.data.models import TKR
from tierkreis.controller.executor.uv_executor import UvExecutor
from tierkreis.controller.storage.filestorage import ControllerFileStorage
from tests.errors.failing_worker.stubs import fail, wont_fail, exit_code_1


def will_fail_graph():
    graph = GraphBuilder(EmptyModel, TKR[int])
    graph.outputs(graph.task(fail()))
    return graph


def wont_fail_graph():
    graph = GraphBuilder(EmptyModel, TKR[int])
    graph.outputs(graph.task(wont_fail()))
    return graph


def fail_in_eval():
    graph = GraphBuilder(EmptyModel, TKR[int])
    graph.outputs(graph.eval(will_fail_graph(), EmptyModel()))
    return graph


def non_zero_exit_code():
    graph = GraphBuilder(EmptyModel, TKR[int])
    graph.outputs(graph.task(exit_code_1()))
    return graph


def test_raise_error() -> None:
    g = will_fail_graph()
    storage = ControllerFileStorage(UUID(int=42), name="will_fail")
    executor = UvExecutor(Path(__file__).parent, logs_path=storage.logs_path)
    storage.clean_graph_files()
    run_graph(storage, executor, g.get_data(), {}, n_iterations=1000)
    assert storage.node_has_error(Loc("-.N0"))


def test_raises_no_error() -> None:
    g = wont_fail_graph()
    storage = ControllerFileStorage(UUID(int=43), name="wont_fail")
    executor = UvExecutor(Path(__file__).parent, logs_path=storage.logs_path)
    storage.clean_graph_files()
    run_graph(storage, executor, g.get_data(), {}, n_iterations=100)
    assert not storage.node_has_error(Loc("-.N0"))


def test_nested_error() -> None:
    g = fail_in_eval()
    storage = ControllerFileStorage(UUID(int=44), name="eval_will_fail")
    executor = UvExecutor(Path(__file__).parent, logs_path=storage.logs_path)
    storage.clean_graph_files()
    run_graph(storage, executor, g.get_data(), {}, n_iterations=1000)
    assert (storage.logs_path.parent / "-/errors").exists()


def test_non_zero_exit_code() -> None:
    g = non_zero_exit_code()
    storage = ControllerFileStorage(UUID(int=46), name="non_zero_exit_code")
    executor = UvExecutor(Path(__file__).parent, logs_path=storage.logs_path)
    storage.clean_graph_files()
    run_graph(storage, executor, g.get_data(), {}, n_iterations=1000)
    assert (storage.logs_path.parent / "-/_error").exists()



from pathlib import Path

from tierkreis.controller.executor.protocol import ControllerExecutor
from tierkreis.exceptions import TierkreisError


class MultipleExecutor:
    """Composes multiple executors into a single object.

    Implements: :py:class:`tierkreis.controller.executor.protocol.ControllerExecutor`
    """

    def __init__(
        self,
        default: ControllerExecutor,
        executors: dict[str, ControllerExecutor],
        assignments: dict[str, str],
    ) -> None:
        self.default = default
        self.executors = executors
        self.assignments = assignments

    def run(self, launcher_name: str, worker_call_args_path: Path) -> None:
        executor_name = self.assignments.get(launcher_name, None)
        # If there is no assignment for the worker, use the default.
        if executor_name is None:
            return self.default.run(launcher_name, worker_call_args_path)

        executor = self.executors.get(executor_name)
        if executor is None:
            raise TierkreisError(
                f"{launcher_name} is assigned to non-existent executor name: {executor_name}."
            )

        return executor.run(launcher_name, worker_call_args_path)



"""Code generated from nexus_worker namespace. Please do not edit."""

from typing import NamedTuple, Union
from tierkreis.controller.data.models import TKR, OpaqueType


class upload_circuit(NamedTuple):
    project_name: TKR[str]  # noqa: F821 # fmt: skip
    circ: TKR[OpaqueType["pytket._tket.circuit.Circuit"]]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[Union[OpaqueType["qnexus.models.references.CircuitRef"], OpaqueType["qnexus.models.references.HUGRRef"]]]]:  # noqa: F821 # fmt: skip
        return TKR[Union[OpaqueType["qnexus.models.references.CircuitRef"], OpaqueType["qnexus.models.references.HUGRRef"]]]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "nexus_worker"


class start_execute_job(NamedTuple):
    project_name: TKR[str]  # noqa: F821 # fmt: skip
    job_name: TKR[str]  # noqa: F821 # fmt: skip
    circuits: TKR[list[Union[OpaqueType["qnexus.models.references.CircuitRef"], OpaqueType["qnexus.models.references.HUGRRef"]]]]  # noqa: F821 # fmt: skip
    n_shots: TKR[list[int]]  # noqa: F821 # fmt: skip
    backend_config: TKR[Union[OpaqueType["quantinuum_schemas.models.backend_config.AerConfig"], OpaqueType["quantinuum_schemas.models.backend_config.AerStateConfig"], OpaqueType["quantinuum_schemas.models.backend_config.AerUnitaryConfig"], OpaqueType["quantinuum_schemas.models.backend_config.BraketConfig"], OpaqueType["quantinuum_schemas.models.backend_config.QuantinuumConfig"], OpaqueType["quantinuum_schemas.models.backend_config.IBMQConfig"], OpaqueType["quantinuum_schemas.models.backend_config.IBMQEmulatorConfig"], OpaqueType["quantinuum_schemas.models.backend_config.ProjectQConfig"], OpaqueType["quantinuum_schemas.models.backend_config.QulacsConfig"], OpaqueType["quantinuum_schemas.models.backend_config.SeleneQuestConfig"], OpaqueType["quantinuum_schemas.models.backend_config.SeleneStimConfig"], OpaqueType["quantinuum_schemas.models.backend_config.SeleneLeanConfig"], OpaqueType["quantinuum_schemas.models.backend_config.SeleneCoinflipConfig"], OpaqueType["quantinuum_schemas.models.backend_config.SeleneClassicalReplayConfig"]]]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[OpaqueType["qnexus.models.references.ExecuteJobRef"]]]:  # noqa: F821 # fmt: skip
        return TKR[OpaqueType["qnexus.models.references.ExecuteJobRef"]]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "nexus_worker"


class is_running(NamedTuple):
    execute_ref: TKR[OpaqueType["qnexus.models.references.ExecuteJobRef"]]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[bool]]:  # noqa: F821 # fmt: skip
        return TKR[bool]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "nexus_worker"


class get_results(NamedTuple):
    execute_ref: TKR[OpaqueType["qnexus.models.references.ExecuteJobRef"]]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[list[OpaqueType["pytket.backends.backendresult.BackendResult"]]]]:  # noqa: F821 # fmt: skip
        return TKR[list[OpaqueType["pytket.backends.backendresult.BackendResult"]]]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "nexus_worker"


class check_status(NamedTuple):
    execute_ref: TKR[OpaqueType["qnexus.models.references.ExecuteJobRef"]]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[str]]:  # noqa: F821 # fmt: skip
        return TKR[str]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "nexus_worker"


class submit(NamedTuple):
    circuits: TKR[list[OpaqueType["pytket._tket.circuit.Circuit"]]]  # noqa: F821 # fmt: skip
    n_shots: TKR[int]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[OpaqueType["qnexus.models.references.ExecuteJobRef"]]]:  # noqa: F821 # fmt: skip
        return TKR[OpaqueType["qnexus.models.references.ExecuteJobRef"]]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "nexus_worker"



"""Code generated from pytket_worker namespace. Please do not edit."""

from typing import NamedTuple, Union
from types import NoneType
from tierkreis.controller.data.models import TKR, OpaqueType


class add_measure_all(NamedTuple):
    circuit: TKR[OpaqueType["pytket._tket.circuit.Circuit"]]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[OpaqueType["pytket._tket.circuit.Circuit"]]]:  # noqa: F821 # fmt: skip
        return TKR[OpaqueType["pytket._tket.circuit.Circuit"]]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "pytket_worker"


class append_pauli_measurement_impl(NamedTuple):
    circuit: TKR[OpaqueType["pytket._tket.circuit.Circuit"]]  # noqa: F821 # fmt: skip
    pauli_string: TKR[OpaqueType["pytket._tket.pauli.QubitPauliString"]]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[OpaqueType["pytket._tket.circuit.Circuit"]]]:  # noqa: F821 # fmt: skip
        return TKR[OpaqueType["pytket._tket.circuit.Circuit"]]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "pytket_worker"


class optimise_phase_gadgets(NamedTuple):
    circuit: TKR[OpaqueType["pytket._tket.circuit.Circuit"]]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[OpaqueType["pytket._tket.circuit.Circuit"]]]:  # noqa: F821 # fmt: skip
        return TKR[OpaqueType["pytket._tket.circuit.Circuit"]]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "pytket_worker"


class apply_pass(NamedTuple):
    circuit: TKR[OpaqueType["pytket._tket.circuit.Circuit"]]  # noqa: F821 # fmt: skip
    compiler_pass: TKR[OpaqueType["pytket._tket.passes.BasePass"]]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[OpaqueType["pytket._tket.circuit.Circuit"]]]:  # noqa: F821 # fmt: skip
        return TKR[OpaqueType["pytket._tket.circuit.Circuit"]]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "pytket_worker"


class compile(NamedTuple):
    circuit: TKR[Union[OpaqueType["pytket._tket.circuit.Circuit"], str, bytes]]  # noqa: F821 # fmt: skip
    input_format: TKR[str] | None = None  # noqa: F821 # fmt: skip
    optimization_level: TKR[int] | None = None  # noqa: F821 # fmt: skip
    gate_set: TKR[Union[list[str], NoneType]] | None = None  # noqa: F821 # fmt: skip
    coupling_map: TKR[Union[list[tuple[int, int]], NoneType]] | None = None  # noqa: F821 # fmt: skip
    output_format: TKR[str] | None = None  # noqa: F821 # fmt: skip
    optimization_pass: TKR[Union[OpaqueType["pytket._tket.passes.BasePass"], NoneType]] | None = None  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[Union[OpaqueType["pytket._tket.circuit.Circuit"], str, bytes]]]:  # noqa: F821 # fmt: skip
        return TKR[Union[OpaqueType["pytket._tket.circuit.Circuit"], str, bytes]]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "pytket_worker"


class compile_circuit_quantinuum(NamedTuple):
    circuit: TKR[OpaqueType["pytket._tket.circuit.Circuit"]]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[OpaqueType["pytket._tket.circuit.Circuit"]]]:  # noqa: F821 # fmt: skip
        return TKR[OpaqueType["pytket._tket.circuit.Circuit"]]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "pytket_worker"


class compile_circuits_quantinuum(NamedTuple):
    circuits: TKR[list[OpaqueType["pytket._tket.circuit.Circuit"]]]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[list[OpaqueType["pytket._tket.circuit.Circuit"]]]]:  # noqa: F821 # fmt: skip
        return TKR[list[OpaqueType["pytket._tket.circuit.Circuit"]]]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "pytket_worker"


class compile_tket_circuit_ibm(NamedTuple):
    circuit: TKR[OpaqueType["pytket._tket.circuit.Circuit"]]  # noqa: F821 # fmt: skip
    backend_name: TKR[str]  # noqa: F821 # fmt: skip
    optimization_level: TKR[int] | None = None  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[OpaqueType["pytket._tket.circuit.Circuit"]]]:  # noqa: F821 # fmt: skip
        return TKR[OpaqueType["pytket._tket.circuit.Circuit"]]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "pytket_worker"


class compile_tket_circuits_ibm(NamedTuple):
    circuits: TKR[list[OpaqueType["pytket._tket.circuit.Circuit"]]]  # noqa: F821 # fmt: skip
    backend_name: TKR[str]  # noqa: F821 # fmt: skip
    optimization_level: TKR[int] | None = None  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[list[OpaqueType["pytket._tket.circuit.Circuit"]]]]:  # noqa: F821 # fmt: skip
        return TKR[list[OpaqueType["pytket._tket.circuit.Circuit"]]]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "pytket_worker"


class compile_tket_circuit_quantinuum(NamedTuple):
    circuit: TKR[OpaqueType["pytket._tket.circuit.Circuit"]]  # noqa: F821 # fmt: skip
    backend_name: TKR[str]  # noqa: F821 # fmt: skip
    optimization_level: TKR[int] | None = None  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[OpaqueType["pytket._tket.circuit.Circuit"]]]:  # noqa: F821 # fmt: skip
        return TKR[OpaqueType["pytket._tket.circuit.Circuit"]]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "pytket_worker"


class compile_tket_circuits_quantinuum(NamedTuple):
    circuits: TKR[list[OpaqueType["pytket._tket.circuit.Circuit"]]]  # noqa: F821 # fmt: skip
    backend_name: TKR[str]  # noqa: F821 # fmt: skip
    optimization_level: TKR[int] | None = None  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[list[OpaqueType["pytket._tket.circuit.Circuit"]]]]:  # noqa: F821 # fmt: skip
        return TKR[list[OpaqueType["pytket._tket.circuit.Circuit"]]]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "pytket_worker"


class to_qasm_str(NamedTuple):
    circuit: TKR[OpaqueType["pytket._tket.circuit.Circuit"]]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[str]]:  # noqa: F821 # fmt: skip
        return TKR[str]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "pytket_worker"


class from_gasm_str(NamedTuple):
    qasm: TKR[str]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[OpaqueType["pytket._tket.circuit.Circuit"]]]:  # noqa: F821 # fmt: skip
        return TKR[OpaqueType["pytket._tket.circuit.Circuit"]]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "pytket_worker"


class to_qir_bytes(NamedTuple):
    circuit: TKR[OpaqueType["pytket._tket.circuit.Circuit"]]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[bytes]]:  # noqa: F821 # fmt: skip
        return TKR[bytes]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "pytket_worker"


class from_qir_bytes(NamedTuple):
    qir: TKR[bytes]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[OpaqueType["pytket._tket.circuit.Circuit"]]]:  # noqa: F821 # fmt: skip
        return TKR[OpaqueType["pytket._tket.circuit.Circuit"]]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "pytket_worker"


class expectation(NamedTuple):
    backend_result: TKR[OpaqueType["pytket.backends.backendresult.BackendResult"]]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[float]]:  # noqa: F821 # fmt: skip
        return TKR[float]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "pytket_worker"


class n_qubits(NamedTuple):
    circuit: TKR[OpaqueType["pytket._tket.circuit.Circuit"]]  # noqa: F821 # fmt: skip

    @staticmethod
    def out() -> type[TKR[int]]:  # noqa: F821 # fmt: skip
        return TKR[int]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "pytket_worker"



from pathlib import Path
from typing import Protocol

from tierkreis.controller.data.location import WorkerCallArgs


class WorkerStorage(Protocol):
    def resolve(self, path: Path | str) -> Path: ...
    def read_call_args(self, path: Path) -> WorkerCallArgs: ...
    def read_input(self, path: Path) -> bytes: ...
    def write_output(self, path: Path, value: bytes) -> None: ...
    def glob(self, path_string: str) -> list[str]: ...
    def mark_done(self, path: Path) -> None: ...
    def write_error(self, path: Path, error_logs: str) -> None: ...



"""Types that are allowed in the Tierkreis worker specification IDL.

We use https://typespec.io/docs/language-basics/built-in-types/ as a guide.
"""

from types import NoneType
from typing import ForwardRef
from tierkreis.idl.models import GenericType
from tierkreis.idl.parser import Parser, lit, reg, seq

type _TypeT = type | ForwardRef


signed_int = lit("integer", "int64", "int32", "int16", "int8", "safeint")
unsigned_int = lit("uint64", "uint32", "uint16", "uint8")
integer_t = (signed_int | unsigned_int).coerce(GenericType(int, []))
float_t = lit("float", "float32", "float64", "numeric").coerce(GenericType(float, []))
bytes_t = lit("bytes").coerce(GenericType(bytes, []))
string_t = lit("string", "url").coerce(GenericType(str, []))
bool_t = lit("bool").coerce(GenericType(bool, []))
none_t = lit("null").coerce(GenericType(NoneType, []))
decimal_t = lit("decimal", "decimal128").fail("Decimal")
plain_datetime = lit("plainDate", "plainTime")
other_date = lit("utcDateTime", "offsetDateTime", "duration")
date_t = (plain_datetime | other_date).fail("Date")
unknown_t = lit("unknown", "void", "never").fail("Unknown")
ident = reg(r"[a-zA-Z0-9_]+")
forward_ref = ident.map(ForwardRef)
generics = (lit("<") >> ident.rep(lit(",")) << lit(">")).opt().map(lambda x: x or [])


def array_t(ins: str) -> tuple[GenericType, str]:
    return (lit("Array<") >> type_symbol << lit(">")).map(
        lambda x: GenericType(list, [x])
    )(ins)


def record_t(ins: str) -> tuple[GenericType, str]:
    return (lit("Record<") >> type_symbol << lit(">")).map(
        lambda x: GenericType(dict, [GenericType(str, []), x])
    )(ins)


@Parser
def generic_t(ins: str) -> tuple[GenericType, str]:
    return seq(
        ident,
        (lit("<") >> ident.rep(lit(",")) << lit(">")).opt().map(lambda x: x or []),
    ).map(lambda x: GenericType(x[0], x[1]))(ins)


@Parser
def type_symbol(ins: str) -> tuple[GenericType, str]:
    return (
        integer_t
        | float_t
        | bytes_t
        | bool_t
        | none_t
        | string_t
        | decimal_t
        | unknown_t
        | date_t
        | array_t
        | record_t
        | generic_t
    )(ins)



from dataclasses import dataclass
from datetime import datetime
from types import NoneType, UnionType
from typing import Mapping, Sequence, TypeVar
from uuid import UUID
from pydantic import BaseModel
import pytest
from tierkreis.controller.data.types import (
    PType,
    bytes_from_ptype,
    generics_in_ptype,
    is_ptype,
    ptype_from_bytes,
)


class UntupledModel[U, V](BaseModel):
    a: U
    b: V


class DummyBaseModel(BaseModel):
    a: int
    b: bytes


@dataclass
class DummyDictConvertible:
    a: int

    def to_dict(self) -> dict:
        return {"a": self.a}

    @classmethod
    def from_dict(cls, args: dict) -> "DummyDictConvertible":
        return DummyDictConvertible(a=args["a"])


@dataclass
class DummyListConvertible:
    a: int

    def to_list(self) -> list:
        return list(range(self.a))

    @classmethod
    def from_list(cls, args: list) -> "DummyListConvertible":
        return DummyListConvertible(a=len(args))


type_list: Sequence[type[PType] | UnionType] = []
type_list.append(bool)
type_list.append(int)
type_list.append(str)
type_list.append(float)
type_list.append(NoneType)
type_list.append(bytes)
type_list.append(Sequence[str])
type_list.append(Sequence[Sequence[Sequence[NoneType]]])
type_list.append(Sequence[str | Sequence[int]])
type_list.append(Sequence[str | Sequence[int | float]])
type_list.append(int | None)
type_list.append(int | bytes)
type_list.append(tuple[int, str])
type_list.append(tuple[int | str])
type_list.append(Mapping[str, dict[str, bytes]])

fail_list: Sequence[type] = []
fail_list.append(UUID)
fail_list.append(datetime)
fail_list.append(list[datetime])
fail_list.append(Mapping[str, dict[str, datetime]])

ptypes: Sequence[PType] = [
    True,
    False,
    0,
    1,
    2,
    "0",
    "1",
    "test",
    0.0,
    0.1,
    15.2,
    None,
    ["one", 1],
    [1, 2, 3],
    [[[]], 1, ["45", "test", b"test bytes"]],
    b"these are some test bytes",
    {"one": 1, "two": "two", "bytes": b"asdf"},
    DummyDictConvertible(a=11),
    DummyListConvertible(a=5),
    DummyBaseModel(a=5, b=b"some bytes"),
]


@pytest.mark.parametrize("ptype", ptypes)
def test_bytes_roundtrip(ptype: PType):
    bs = bytes_from_ptype(ptype)
    new_type = ptype_from_bytes(bs, type(ptype))
    assert ptype == new_type


annotated_ptypes: Sequence[tuple[PType, type]] = [
    (ptype, type(ptype)) for ptype in ptypes
] + [  # Not possible to deserialise without annotations
    (
        [DummyListConvertible(a=1), DummyListConvertible(a=2)],
        list[DummyListConvertible],
    ),
    (
        {"a": DummyListConvertible(a=1), "b": DummyListConvertible(a=2)},
        dict[str, DummyListConvertible],
    ),
]


@pytest.mark.parametrize("annotated_ptype", annotated_ptypes)
def test_annotated_bytes_roundtrip(annotated_ptype: tuple[PType, type]):
    ptype, annotation = annotated_ptype
    bs = bytes_from_ptype(ptype)
    new_type = ptype_from_bytes(bs, annotation)
    assert ptype == new_type


@pytest.mark.parametrize("ptype", type_list)
def test_ptype_from_annotation(ptype: type[PType]):
    assert is_ptype(ptype)


@pytest.mark.parametrize("ptype", fail_list)
def test_ptype_from_annotation_fails(ptype: type[PType]):
    assert not is_ptype(ptype)


S = TypeVar("S")
T = TypeVar("T")

generic_types = []
generic_types.append((list[T], {str(T)}))  # type: ignore
generic_types.append((list[S | T], {str(S), str(T)}))  # type: ignore
generic_types.append((list[list[list[T]]], {str(T)}))  # type: ignore
generic_types.append((tuple[S, T], {str(S), str(T)}))  # type: ignore
generic_types.append((UntupledModel[S, T], {str(S), str(T)}))  # type: ignore


@pytest.mark.parametrize("ptype,generics", generic_types)
def test_generic_types(ptype: type[PType], generics: set[type[PType]]):
    assert generics_in_ptype(ptype) == generics



from pathlib import Path
import platform
from pydantic import BaseModel, Field


class MpiSpec(BaseModel):
    max_proc_per_node: int | None = 1
    proc: int | None = None


class ResourceSpec(BaseModel):
    nodes: int = 1
    cores_per_node: int | None = 1
    memory_gb: int | None = 4
    gpus_per_node: int | None = 0


class UserSpec(BaseModel):
    mail: str | None = None  # some clusters require this


class ContainerSpec(BaseModel):
    image: str
    engine: str  # e.g. singularity, docker, enroot?
    name: str | None = None
    extra_args: dict[str, str | None] = Field(default_factory=dict)
    env_file: str | None = None


class JobSpec(BaseModel):
    job_name: str
    command: str  # used instead of popen.input
    resource: ResourceSpec
    account: str | None = None
    """Account or group used to submit this job."""
    mpi: MpiSpec | None = None
    user: UserSpec | None = None
    container: ContainerSpec | None = None
    walltime: str = "01:00:00"  # HH:MM:SS Replacement?
    queue: str | None = None
    output_path: Path | None = None
    error_path: Path | None = None
    extra_scheduler_args: dict[str, str | None] = Field(default_factory=dict)
    environment: dict[str, str] = Field(default_factory=dict)
    include_no_check_directory_flag: bool = False


def pjsub_large_spec() -> JobSpec:
    arch = platform.machine()
    uv_path = Path.home() / ".local" / f"bin_{arch}" / "uv"
    return JobSpec(
        job_name="pjsub_large",
        account="hp240496",
        command=f"{str(uv_path)} run main.py",
        queue="q-QTM-M",
        resource=ResourceSpec(nodes=32),
        environment={
            "VIRTUAL_ENVIRONMENT": "",
            "UV_PROJECT_ENVIRONMENT": f"venv_{arch}",
            "PJM_LLIO_GFSCACHE": "/vol0004",
        },
        walltime="03:00:00",
        extra_scheduler_args={
            "-z": "jid",
            "--llio": "sharedtmp-size=64Gi",
        },
        mpi=MpiSpec(max_proc_per_node=4),
    )


def pjsub_small_spec() -> JobSpec:
    arch = platform.machine()
    uv_path = Path.home() / ".local" / f"bin_{arch}" / "uv"
    return JobSpec(
        job_name="pjsub_small",
        account="hp240496",
        command=f"{str(uv_path)} run main.py",
        resource=ResourceSpec(nodes=1),
        environment={
            "VIRTUAL_ENVIRONMENT": "",
            "UV_PROJECT_ENVIRONMENT": f"venv_{arch}",
        },
        walltime="00:15:00",
        extra_scheduler_args={
            "-z": "jid",
            "--llio": "sharedtmp-size=10Gi",
        },
        mpi=MpiSpec(max_proc_per_node=4),
    )



from __future__ import annotations

import argparse
import importlib
import json
import logging
import sys
from pathlib import Path
from typing import Any, Callable

from tierkreis.cli.run_workflow import run_workflow
from tierkreis.controller.data.graph import GraphData
from tierkreis.controller.data.types import PType, ptype_from_bytes
from tierkreis.exceptions import TierkreisError


def _import_from_path(module_name: str, file_path: str) -> Any:
    spec = importlib.util.spec_from_file_location(module_name, file_path)  # type: ignore
    module = importlib.util.module_from_spec(spec)  # type: ignore
    sys.modules[module_name] = module
    spec.loader.exec_module(module)
    return module


def load_graph(graph_input: str) -> GraphData:
    if ":" not in graph_input:
        raise TierkreisError(f"Invalid argument: {graph_input}")
    module_name, function_name = graph_input.split(":")
    print(f"Loading graph from module '{module_name}' and function '{function_name}'")
    if ".py" in module_name:
        module = _import_from_path("graph_module", module_name)
    else:
        module = importlib.import_module(module_name, __package__)
    build_submission_graph: Callable[[], GraphData] = getattr(module, function_name)

    return build_submission_graph()


def _load_inputs(input_files: list[str]) -> dict[str, PType]:
    if len(input_files) == 1 and input_files[0].endswith(".json"):
        with open(input_files[0], "r") as fh:
            return {k: json.dumps(v).encode() for k, v in json.load(fh).items()}
    inputs = {}
    for input_file in input_files:
        if ":" not in input_file:
            raise TierkreisError(f"Invalid argument: {input_file}")
        key, value = input_file.split(":")
        with open(value, "rb") as fh:
            inputs[key] = ptype_from_bytes(fh.read())
    return inputs


def parse_args(
    main_parser: argparse._SubParsersAction[argparse.ArgumentParser],
) -> argparse.ArgumentParser:
    parser = main_parser.add_parser(
        name="run",
        description="Tierkreis: a workflow engine for quantum HPC.",
    )
    graph = parser.add_mutually_exclusive_group(required=True)
    graph.add_argument(
        "-f", "--from-file", type=Path, help="Load a graph from a .json file"
    )
    graph.add_argument(
        "-g",
        "--graph-location",
        help="Fully qualifying name of a Callable () -> GraphData. "
        + "Example: tierkreis.cli.sample_graph:simple_eval"
        + "Or a path to a python file and function."
        + "Example: examples/hello_world/hello_world_graph.py:hello_graph",
        type=str,
    )
    parser.add_argument(
        "-i",
        "--input-files",
        nargs="*",
        help="Graph inputs:"
        "Either a single .json file or a key value list  port1:path1 port2:path2"
        + "where path is a binary file.",
    )
    parser.add_argument(
        "--run-id", default=None, type=int, help="Set a workflow run id"
    )
    parser.add_argument("--name", default=None, type=str, help="Set a workflow name")
    parser.add_argument(
        "-l",
        "--loglevel",
        default=logging.WARNING,
        choices=logging.getLevelNamesMapping().keys(),
        help="Set log level.",
    )
    parser.add_argument("-v", "--verbose", action="store_true")
    parser.add_argument(
        "--registry-path", default=None, type=Path, help="Location of executable tasks."
    )
    parser.add_argument(
        "-o",
        "--print-output",
        action="store_true",
        help="Print the outputs of the top-level node. ",
    )

    # run_graph() arguments
    parser.add_argument(
        "-n",
        "--n-iterations",
        default=10**5,
        type=int,
        help="Set the maximum number of iterations.",
    )
    parser.add_argument(
        "-p",
        "--polling-interval-seconds",
        default=0.01,
        type=float,
        help="Set the controller tickrate.",
    )
    parser.add_argument(
        "-r",
        "--do-clean-restart",
        action="store_true",
        help="Clear graph files before running",
    )
    parser.add_argument("--uv", action="store_true", help="Use uv worker")

    return parser


def run_workflow_args(args: argparse.Namespace):
    if args.verbose:
        args.log_level = logging.DEBUG

    if args.graph_location is not None:
        graph = load_graph(args.graph_location)
    else:
        with open(args.from_file, "r") as fh:
            graph = ptype_from_bytes(fh.read().encode(), GraphData)
    if args.input_files is not None:
        inputs = _load_inputs(args.input_files)
    else:
        inputs = {}
    print(inputs)
    run_workflow(
        graph,
        inputs,
        name=args.name,
        run_id=args.run_id,
        registry_path=args.registry_path,
        use_uv_worker=args.uv,
        n_iterations=args.n_iterations,
        polling_interval_seconds=args.polling_interval_seconds,
        print_output=args.print_output,
    )


class TierkreisCli:
    @staticmethod
    def add_subcommand(
        main_parser: argparse._SubParsersAction[argparse.ArgumentParser],
    ) -> None:
        parser = parse_args(main_parser)
        parser.set_defaults(func=TierkreisCli.execute)

    @staticmethod
    def execute(args: argparse.Namespace) -> None:
        run_workflow_args(args)


def main() -> None:
    parser = argparse.ArgumentParser(
        prog="tkr",
        description="The main tierkreis command-line tool.",
    )
    subparser = parser.add_subparsers(title="subcommands")
    TierkreisCli.add_subcommand(subparser)
    try:
        from tierkreis_visualization.cli import TierkreisVizCli

        TierkreisVizCli.add_subcommand(subparser)
    except ImportError:
        logging.warning("Could not import Tierkreis Visualization CLI")
        logging.warning(
            "To install it, please run 'pip install tierkreis-visualization'"
        )
    args = parser.parse_args(args=None if sys.argv[1:] else ["--help"])
    args.func(args)


if __name__ == "__main__":
    main()



"""Code generated from failing_worker namespace. Please do not edit."""

from typing import NamedTuple
from tierkreis.controller.data.models import TKR


class fail(NamedTuple):
    @staticmethod
    def out() -> type[TKR[int]]:  # noqa: F821 # fmt: skip
        return TKR[int]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "failing_worker"


class wont_fail(NamedTuple):
    @staticmethod
    def out() -> type[TKR[int]]:  # noqa: F821 # fmt: skip
        return TKR[int]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "failing_worker"


class exit_code_1(NamedTuple):
    @staticmethod
    def out() -> type[TKR[int]]:  # noqa: F821 # fmt: skip
        return TKR[int]  # noqa: F821 # fmt: skip

    @property
    def namespace(self) -> str:
        return "failing_worker"



from typing import NamedTuple
from tierkreis.builtins.stubs import iadd, igt, itimes, tuple_impl, untuple, mod
from tierkreis.controller.data.core import EmptyModel
from tierkreis.builder import GraphBuilder
from tierkreis.controller.data.models import TKR


class DoublerInput(NamedTuple):
    x: TKR[int]
    intercept: TKR[int]


class DoublerOutput(NamedTuple):
    a: TKR[int]
    value: TKR[int]


def typed_doubler_plus_multi():
    g = GraphBuilder(DoublerInput, DoublerOutput)
    mul = g.task(itimes(a=g.inputs.x, b=g.const(2)))
    out = g.task(iadd(a=mul, b=g.inputs.intercept))
    g.outputs(DoublerOutput(a=g.inputs.x, value=out))
    return g


def typed_doubler_plus():
    g = GraphBuilder(DoublerInput, TKR[int])
    mul = g.task(itimes(a=g.inputs.x, b=g.const(2)))
    out = g.task(iadd(a=mul, b=g.inputs.intercept))
    g.outputs(out)
    return g


class TypedEvalOutputs(NamedTuple):
    typed_eval_output: TKR[int]


def typed_eval():
    g = GraphBuilder(EmptyModel, TypedEvalOutputs)
    e = g.eval(typed_doubler_plus(), DoublerInput(x=g.const(6), intercept=g.const(0)))
    g.outputs(TypedEvalOutputs(typed_eval_output=e))
    return g


class LoopBodyInput(NamedTuple):
    loop_acc: TKR[int]


class LoopBodyOutput(NamedTuple):
    loop_acc: TKR[int]
    should_continue: TKR[bool]


def loop_body():
    g = GraphBuilder(LoopBodyInput, LoopBodyOutput)
    a_plus = g.task(iadd(a=g.inputs.loop_acc, b=g.const(1)))
    pred = g.task(igt(a=g.const(10), b=a_plus))
    g.outputs(LoopBodyOutput(loop_acc=a_plus, should_continue=pred))
    return g


class TypedLoopOutput(NamedTuple):
    typed_loop_output: TKR[int]


def typed_loop():
    g = GraphBuilder(EmptyModel, TypedLoopOutput)
    loop = g.loop(loop_body(), LoopBodyInput(loop_acc=g.const(6)))
    g.outputs(TypedLoopOutput(typed_loop_output=loop.loop_acc))
    return g


class TypedMapOutput(NamedTuple):
    typed_map_output: TKR[list[int]]


def typed_map():
    g = GraphBuilder(EmptyModel, TypedMapOutput)
    Ns = g.const(list(range(21)))
    ins = g.map(lambda n: DoublerInput(x=n, intercept=g.const(6)), Ns)
    m = g.map(typed_doubler_plus(), ins)
    g.outputs(TypedMapOutput(typed_map_output=m))
    return g


def typed_destructuring():
    g = GraphBuilder(EmptyModel, TypedMapOutput)
    Ns = g.const(list(range(21)))
    ins = g.map(lambda n: DoublerInput(x=n, intercept=g.const(6)), Ns)
    m = g.map(typed_doubler_plus_multi(), ins)
    mout = g.map(lambda x: x.value, m)
    g.outputs(TypedMapOutput(typed_map_output=mout))
    return g


def tuple_untuple():
    g = GraphBuilder(EmptyModel, TKR[int])
    t = g.task(tuple_impl(g.const(1), g.const(2)))
    ut = g.task(untuple(t))
    g.outputs(g.task(iadd(ut.a, ut.b)))
    return g


def factorial():
    g = GraphBuilder(TKR[int], TKR[int])
    pred = g.task(igt(g.inputs, g.const(1)))
    n_minus_one = g.task(iadd(g.const(-1), g.inputs))
    rec = g.eval(g.ref(), n_minus_one)
    out = g.ifelse(pred, g.task(itimes(g.inputs, rec)), g.const(1))
    g.outputs(out)
    return g


class GCDInput(NamedTuple):
    a: TKR[int]
    b: TKR[int]


def gcd():
    g = GraphBuilder(GCDInput, TKR[int])

    pred = g.task(igt(g.inputs.b, g.const(0)))
    a_mod_b = g.task(mod(g.inputs.a, g.inputs.b))
    rec = g.eval(g.ref(), GCDInput(a=g.inputs.b, b=a_mod_b))

    g.outputs(g.ifelse(pred, rec, g.inputs.a))
    return g



from typing import Generic, NamedTuple, TypeVar
from tierkreis.builder import GraphBuilder, TypedGraphRef
from tierkreis.builtins.stubs import head, igt, impl_len
from tierkreis.controller.data.graph import GraphData
from tierkreis.controller.data.models import TKR
from tierkreis.controller.data.types import PType


class FoldGraphOuterInputs[A: PType, B: PType](NamedTuple):
    func: TKR[GraphData]
    accum: TKR[B]
    values: TKR[list[A]]


class FoldGraphOuterOutputs[A: PType, B: PType](NamedTuple):
    accum: TKR[B]
    values: TKR[list[A]]
    should_continue: TKR[bool]


class InnerFuncInput[A: PType, B: PType](NamedTuple):
    accum: TKR[B]
    value: TKR[A]


def _fold_graph_outer[A: PType, B: PType]():
    g = GraphBuilder(FoldGraphOuterInputs[A, B], FoldGraphOuterOutputs[A, B])

    func = g.inputs.func
    accum = g.inputs.accum
    values = g.inputs.values

    values_len = g.task(impl_len(values))
    # True if there is more than one value in the list.
    non_empty = g.task(igt(values_len, g.const(0)))

    # Will only succeed if values is non-empty.
    headed = g.task(head(values))

    # Apply the function if we were able to pop off a value.
    tgd = TypedGraphRef[InnerFuncInput, TKR[B]](
        func.value_ref(), TKR[B], InnerFuncInput
    )
    applied_next = g.eval(tgd, InnerFuncInput(accum, headed.head))

    next_accum = g.ifelse(non_empty, applied_next, accum)
    next_values = g.ifelse(non_empty, headed.rest, values)
    g.outputs(FoldGraphOuterOutputs(next_accum, next_values, non_empty))
    return g


A = TypeVar("A", bound=PType, covariant=True)
B = TypeVar("B", bound=PType, covariant=True)


class FoldGraphInputs(NamedTuple, Generic[A, B]):
    initial: TKR[B]
    values: TKR[list[A]]


class FoldFunctionInput(NamedTuple, Generic[A, B]):
    accum: TKR[B]
    value: TKR[A]


# fold : {func: (b -> a -> b)} -> {initial: b} -> {values: list[a]} -> {value: b}
# fold : { A x B -> B } -> { list[A] x B -> B }
def fold_graph(
    func: GraphBuilder[FoldFunctionInput[A, B], TKR[B]],
) -> GraphBuilder[FoldGraphInputs[A, B], TKR[B]]:
    g = GraphBuilder(FoldGraphInputs[A, B], TKR[B])
    foldfunc = g._graph_const(func)
    # TODO: include the computation inside the fold
    ins = FoldGraphOuterInputs(
        TKR(*foldfunc.graph_ref), g.inputs.initial, g.inputs.values
    )
    loop = g.loop(_fold_graph_outer(), ins)
    g.outputs(loop.accum)
    return g



from logging import getLogger
from pathlib import Path
from random import randint
import statistics
from sys import argv
from time import sleep
from typing import NamedTuple, Sequence

from tierkreis.controller.data.location import WorkerCallArgs
from tierkreis.controller.data.models import portmapping
from tierkreis.controller.data.types import PType, bytes_from_ptype, ptype_from_bytes
from tierkreis.worker.worker import TierkreisWorkerError
from tierkreis.worker.storage.protocol import WorkerStorage
from tierkreis import Worker


logger = getLogger(__name__)

worker = Worker("builtins")


@worker.task()
def iadd(a: int, b: int) -> int:
    logger.debug(f"iadd {a} {b}")
    return a + b


@worker.task()
def itimes(a: int, b: int) -> int:
    logger.debug(f"itimes {a} {b}")
    return a * b


@worker.task()
def igt(a: int, b: int) -> bool:
    logger.debug(f"igt {a} {b}")
    return a > b


@worker.task()
def neg(a: bool) -> bool:
    return not a


@worker.task()
def impl_and(a: bool, b: bool) -> bool:
    logger.debug(f"and {a} {b}")
    return a and b


@worker.task()
def impl_id[T: PType](value: T) -> T:
    logger.debug(f"id {value}")
    return value


@worker.task()
def append[T](v: list[T], a: T) -> list[T]:  # noqa: E741
    v.append(a)
    return v


@portmapping
class Headed[T: PType](NamedTuple):
    head: T
    rest: list[T]


@worker.task()
def head[T: PType](v: list[T]) -> Headed[T]:  # noqa: E741
    head, rest = v[0], v[1:]
    return Headed(head=head, rest=rest)


@worker.task()
def impl_len[A](v: list[A]) -> int:
    logger.info("len: %s", v)
    return len(v)


@worker.task()
def str_eq(a: str, b: str) -> bool:
    return a == b


@worker.task()
def str_neq(a: str, b: str) -> bool:
    return a != b


@worker.primitive_task()
def fold_values(args: WorkerCallArgs, storage: WorkerStorage) -> None:
    values_glob = storage.glob(str(args.inputs["values_glob"]))
    values_glob.sort(key=lambda x: int(Path(x).name.split("-")[-1]))
    bs = [storage.read_input(Path(value)) for value in values_glob]
    values = [ptype_from_bytes(b) for b in bs]
    storage.write_output(Path(args.outputs["value"]), bytes_from_ptype(values))


@worker.primitive_task()
def unfold_values(args: WorkerCallArgs, storage: WorkerStorage) -> None:
    value_list = ptype_from_bytes(storage.read_input(args.inputs["value"]))
    match value_list:
        case list() | Sequence():
            for i, v in enumerate(value_list):
                storage.write_output(args.output_dir / str(i), bytes_from_ptype(v))
        case _:
            raise TierkreisWorkerError(f"Expected list found {value_list}")


@worker.task()
def concat(lhs: str, rhs: str) -> str:
    return lhs + rhs


@worker.task()
def zip_impl[U, V](a: list[U], b: list[V]) -> list[tuple[U, V]]:
    return list(zip(a, b))


@portmapping
class Unzipped[U: PType, V: PType](NamedTuple):
    a: list[U]
    b: list[V]


@worker.task()
def unzip[U: PType, V: PType](value: list[tuple[U, V]]) -> Unzipped[U, V]:
    value_a, value_b = map(list, zip(*value))
    return Unzipped(a=value_a, b=value_b)


@worker.task()
def tuple_impl[U, V](a: U, b: V) -> tuple[U, V]:
    return (a, b)


@portmapping
class Untupled[U: PType, V: PType](NamedTuple):
    a: U
    b: V


@worker.task()
def untuple[U: PType, V: PType](value: tuple[U, V]) -> Untupled[U, V]:
    logger.info("untuple: %s", value)
    value_a, value_b = value
    return Untupled(a=value_a, b=value_b)


@worker.task()
def mean(values: list[float]) -> float:
    return statistics.mean(values)


@worker.task()
def mod(a: int, b: int) -> int:
    return a % b


@worker.task()
def rand_int(a: int, b: int) -> int:
    return randint(a, b)


@worker.task()
def tkr_sleep(delay_seconds: float) -> bool:
    sleep(delay_seconds)
    return True


@worker.task()
def tkr_encode(string: str) -> bytes:
    return string.encode()


@worker.task()
def tkr_decode(bytes: bytes) -> str:
    return bytes.decode()


if __name__ == "__main__":
    worker.app(argv)



from tierkreis.builder import GraphBuilder
from tierkreis.controller.data.graph import GraphData
from tierkreis.controller.data.location import Loc
from tierkreis.controller.data.types import PType, ptype_from_bytes
from tierkreis.controller.storage.protocol import ControllerStorage
from tierkreis.controller.storage.filestorage import (
    ControllerFileStorage as FileStorage,
)
from tierkreis.controller.storage.in_memory import (
    ControllerInMemoryStorage as InMemoryStorage,
)

__all__ = ["FileStorage", "InMemoryStorage"]


def read_outputs(
    g: GraphData | GraphBuilder, storage: ControllerStorage
) -> dict[str, PType] | PType:
    if isinstance(g, GraphBuilder):
        g = g.get_data()

    out_ports = list(g.nodes[g.output_idx()].inputs.keys())
    if len(out_ports) == 1 and "value" in out_ports:
        return ptype_from_bytes(storage.read_output(Loc(), "value"))
    return {k: ptype_from_bytes(storage.read_output(Loc(), k)) for k in out_ports}



from typing import NamedTuple
from tierkreis.controller.data.models import portmapping
from tierkreis import Worker
from tierkreis.controller.data.types import PType

worker = Worker("TestNamespace")


@portmapping
class A(NamedTuple):
    name: dict[str, str]
    age: int


class B(NamedTuple):
    name: dict[str, str]
    age: int


class C[T: PType](NamedTuple):
    a: list[int]
    b: B
    t: T


@worker.task()
def foo(a: int, b: str) -> A: ...
@worker.task()
def bar() -> B: ...
@worker.task()
def z[T: PType](c: C[T]) -> C[T]: ...


expected_namespace = worker.namespace



import logging
from time import sleep

from tierkreis.builder import GraphBuilder
from tierkreis.controller.data.graph import Eval, GraphData
from tierkreis.controller.data.location import Loc
from tierkreis.controller.data.types import PType, bytes_from_ptype, ptype_from_bytes
from tierkreis.controller.executor.protocol import ControllerExecutor
from tierkreis.controller.start import NodeRunData, start, start_nodes
from tierkreis.controller.storage.protocol import ControllerStorage
from tierkreis.controller.storage.walk import walk_node
from tierkreis.controller.data.core import PortID, ValueRef

root_loc = Loc("")
logger = logging.getLogger(__name__)


def run_graph(
    storage: ControllerStorage,
    executor: ControllerExecutor,
    g: GraphData | GraphBuilder,
    graph_inputs: dict[str, PType] | PType,
    n_iterations: int = 10000,
    polling_interval_seconds: float = 0.01,
) -> None:
    if isinstance(g, GraphBuilder):
        g = g.get_data()

    if not isinstance(graph_inputs, dict):
        graph_inputs = {"value": graph_inputs}
    remaining_inputs = g.remaining_inputs({k for k in graph_inputs.keys()})
    if len(remaining_inputs) > 0:
        logger.warning(f"Some inputs were not provided: {remaining_inputs}")

    storage.write_metadata(Loc(""))
    for name, value in graph_inputs.items():
        storage.write_output(root_loc.N(-1), name, bytes_from_ptype(value))

    storage.write_output(root_loc.N(-1), "body", bytes_from_ptype(g))

    inputs: dict[PortID, ValueRef] = {
        k: (-1, k) for k, _ in graph_inputs.items() if k != "body"
    }
    node_run_data = NodeRunData(Loc(), Eval((-1, "body"), inputs), [])
    start(storage, executor, node_run_data)
    resume_graph(storage, executor, n_iterations, polling_interval_seconds)


def resume_graph(
    storage: ControllerStorage,
    executor: ControllerExecutor,
    n_iterations: int = 10000,
    polling_interval_seconds: float = 0.01,
) -> None:
    message = storage.read_output(Loc().N(-1), "body")
    graph = ptype_from_bytes(message, GraphData)

    for _ in range(n_iterations):
        walk_results = walk_node(storage, Loc(), graph.output_idx(), graph)
        if walk_results.errored != []:
            # TODO: add to base class after storage refactor
            (storage.logs_path.parent / "-" / "_error").touch()
            node_errors = "\n".join(x for x in walk_results.errored)
            storage.write_node_errors(Loc(), node_errors)

            print("\n\nGraph finished with errors.\n\n")

            for error_loc in walk_results.errored:
                print(error_loc)
                print(storage.read_errors(error_loc))
                print("\n\n")

            print("--- Tierkreis graph errors above this line. ---\n\n")
            break

        start_nodes(storage, executor, walk_results.inputs_ready)
        if storage.is_node_finished(Loc()):
            break
        sleep(polling_interval_seconds)



from types import NoneType
from typing import NamedTuple
import pytest
from tierkreis.controller.data.models import PModel, dict_from_pmodel, portmapping
from tierkreis.controller.data.types import PType
from tests.controller.test_types import ptypes


@portmapping
class NamedPModel(NamedTuple):
    a: bool
    b: int
    c: float
    d: str
    e: NoneType
    f: list[str]
    g: bytes
    h: tuple[int, str]
    i: list[list[int] | int]


@pytest.mark.parametrize("pmodel", ptypes)
def test_dict_from_pmodel_unnested(pmodel: PModel):
    assert dict_from_pmodel(pmodel) == {"value": pmodel}


named_p_model = NamedPModel(
    a=True,
    b=5,
    c=56.7,
    d="test",
    e=None,
    f=["one", "two", "three"],
    g=b"test bytes",
    h=(5, "test"),
    i=[[], 2],
)
named_p_model_expected = {
    "a": True,
    "b": 5,
    "c": 56.7,
    "d": "test",
    "e": None,
    "f": ["one", "two", "three"],
    "g": b"test bytes",
    "h": (5, "test"),
    "i": [[], 2],
}

pmodels = [(named_p_model, named_p_model_expected)]


@pytest.mark.parametrize("pmodel,expected", pmodels)
def test_dict_from_pmodel_nested(pmodel: PModel, expected: dict[str, PType]):
    assert dict_from_pmodel(pmodel) == expected



# from functools import partial
from functools import partial
from pathlib import Path
from typing import Callable

from tierkreis.controller.executor.hpc.hpc_executor import run_hpc_executor
from tierkreis.controller.executor.hpc.job_spec import (
    JobSpec,
    pjsub_large_spec,
    pjsub_small_spec,
)


_COMMAND_PREFIX = "#PJM"


def generate_pjsub_script(spec: JobSpec) -> str:
    # 1. Shebang and file header
    lines = [
        """#!/bin/bash
#
# PJSUB Job Script generated by TIERKREIS

# --- Core Job Specifications ---"""
    ]
    # 2. Name
    lines.append(f"{_COMMAND_PREFIX} -N {spec.job_name}")
    # 3. Resources (node exclusive)
    lines.append(f'{_COMMAND_PREFIX} -L "elapse={spec.walltime}"')
    lines.append(f'{_COMMAND_PREFIX} -L "node={spec.resource.nodes}"')
    if spec.resource.memory_gb is not None:
        lines.append(f'{_COMMAND_PREFIX} -L "node-mem={spec.resource.memory_gb}"')
    if spec.resource.gpus_per_node is not None:
        lines.append(f'{_COMMAND_PREFIX} -L "gpu={spec.resource.gpus_per_node}"')

    if spec.queue is not None:
        lines.append(f'{_COMMAND_PREFIX} -L "rscqrp={spec.queue}"')
    # 4. User settings
    lines.append("\n# --- User Details ---")
    if spec.account is not None:
        lines.append(f"{_COMMAND_PREFIX} -g {spec.account}")
    if spec.user is not None:
        if spec.user.mail is not None:
            lines.append(f"{_COMMAND_PREFIX} -m e")  # end only
            lines.append(f"{_COMMAND_PREFIX} --mail-list {spec.user.mail}")

    # 5. Output and Error handling
    lines.append("\n# --- Output and Error Handling ---")
    if spec.error_path is not None:
        lines.append(f"{_COMMAND_PREFIX} -e {spec.error_path}")
    else:
        lines.append(f"{_COMMAND_PREFIX} -j")
    if spec.output_path is None:
        spec.output_path = Path(f"./{spec.job_name}.o")
    lines.append(f"{_COMMAND_PREFIX} -o {spec.output_path}")

    # 6. MPI
    if spec.mpi is not None:
        lines.append("\n# --- MPI ---")
        if spec.mpi.proc is not None:
            lines.append(f'{_COMMAND_PREFIX} --mpi "proc={spec.mpi.proc}"')
        if spec.mpi.max_proc_per_node is not None:
            lines.append(
                f'{_COMMAND_PREFIX} --mpi "max-proc-per-node={spec.mpi.max_proc_per_node}"'
            )

    # 7. User specific
    lines.append("\n# --- User-Specific Arguments ---")
    for key, value in spec.extra_scheduler_args.items():
        lines.append(f"{_COMMAND_PREFIX} {key} {value if value is not None else ''}")

    # 8. Environment
    lines.append("\n# --- Environment Setup ---")
    for key, value in spec.environment.items():
        lines.append(f'{_COMMAND_PREFIX} -x "{key}"="{value}"')

    # 9. Container logic

    # 10. User Command, (prologue), command, (epilogue)
    lines.append("\n# --- User Command ---")
    lines.append(spec.command)

    with open("./script", "w+") as fh:
        fh.write("\n".join(lines))

    return "\n".join(lines)


class PJSUBExecutor:
    def __init__(
        self,
        registry_path: Path | None,
        logs_path: Path,
        spec: JobSpec,
        command: str = "pjsub",
    ) -> None:
        self.launchers_path = registry_path
        self.logs_path = logs_path
        self.errors_path = logs_path
        self.spec = spec
        self.script_fn: Callable[[JobSpec], str] = generate_pjsub_script
        self.command = command

    def run(self, launcher_name: str, worker_call_args_path: Path) -> None:
        self.errors_path = (
            self.logs_path.parent.parent / worker_call_args_path.parent / "errors"
        )
        run_hpc_executor(self, launcher_name, worker_call_args_path)


PJSUB_EXECUTOR_SMALL = partial(PJSUBExecutor, spec=pjsub_small_spec())
PJSUB_EXECUTOR_LARGE = partial(PJSUBExecutor, spec=pjsub_large_spec())



import logging
from sys import argv
from tierkreis import Worker

logger = logging.getLogger(__name__)
worker = Worker("failing_worker")


@worker.task()
def fail() -> int:
    logger.error("Raising an error now...")
    raise ValueError("Worker failed!")


@worker.task()
def wont_fail() -> int:
    return 0


@worker.task()
def exit_code_1() -> int:
    exit(1)


if __name__ == "__main__":
    worker.app(argv)



from dataclasses import dataclass, field
import logging
from typing import Any, Callable, Literal, assert_never
from pydantic import BaseModel, RootModel
from tierkreis.controller.data.core import PortID
from tierkreis.controller.data.core import NodeIndex
from tierkreis.controller.data.core import ValueRef
from tierkreis.controller.data.location import Loc, OutputLoc
from tierkreis.controller.data.types import PType, ptype_from_bytes
from tierkreis.exceptions import TierkreisError

logger = logging.getLogger(__name__)


@dataclass
class Func:
    function_name: str
    inputs: dict[PortID, ValueRef]
    outputs: set[PortID] = field(default_factory=lambda: set())
    type: Literal["function"] = field(default="function")


@dataclass
class Eval:
    graph: ValueRef
    inputs: dict[PortID, ValueRef]
    outputs: set[PortID] = field(default_factory=lambda: set())
    type: Literal["eval"] = field(default="eval")


@dataclass
class Loop:
    body: ValueRef
    inputs: dict[PortID, ValueRef]
    continue_port: PortID  # The port that specifies if the loop should continue.
    outputs: set[PortID] = field(default_factory=lambda: set())
    type: Literal["loop"] = field(default="loop")


@dataclass
class Map:
    body: ValueRef
    inputs: dict[PortID, ValueRef]
    outputs: set[PortID] = field(default_factory=lambda: set())
    type: Literal["map"] = field(default="map")


@dataclass
class Const:
    value: Any
    outputs: set[PortID] = field(default_factory=lambda: set())
    inputs: dict[PortID, ValueRef] = field(default_factory=lambda: {})
    type: Literal["const"] = field(default="const")


@dataclass
class IfElse:
    pred: ValueRef
    if_true: ValueRef
    if_false: ValueRef
    outputs: set[PortID] = field(default_factory=lambda: set())
    inputs: dict[PortID, ValueRef] = field(default_factory=lambda: {})
    type: Literal["ifelse"] = field(default="ifelse")


@dataclass
class EagerIfElse:
    pred: ValueRef
    if_true: ValueRef
    if_false: ValueRef
    outputs: set[PortID] = field(default_factory=lambda: set())
    inputs: dict[PortID, ValueRef] = field(default_factory=lambda: {})

    type: Literal["eifelse"] = field(default="eifelse")


@dataclass
class Input:
    name: str
    outputs: set[PortID] = field(default_factory=lambda: set())
    inputs: dict[PortID, ValueRef] = field(default_factory=lambda: {})
    type: Literal["input"] = field(default="input")


@dataclass
class Output:
    inputs: dict[PortID, ValueRef]
    outputs: set[PortID] = field(default_factory=lambda: set())
    type: Literal["output"] = field(default="output")


NodeDef = Func | Eval | Loop | Map | Const | IfElse | EagerIfElse | Input | Output
NodeDefModel = RootModel[NodeDef]


class GraphData(BaseModel):
    nodes: list[NodeDef] = []
    fixed_inputs: dict[PortID, OutputLoc] = {}
    graph_inputs: set[PortID] = set()
    graph_output_idx: NodeIndex | None = None

    def input(self, name: str) -> ValueRef:
        return self.add(Input(name))(name)

    def const(self, value: PType) -> ValueRef:
        return self.add(Const(value))("value")

    def func(
        self, function_name: str, inputs: dict[PortID, ValueRef]
    ) -> Callable[[PortID], ValueRef]:
        return self.add(Func(function_name, inputs))

    def eval(
        self, graph: ValueRef, inputs: dict[PortID, ValueRef]
    ) -> Callable[[PortID], ValueRef]:
        return self.add(Eval(graph, inputs))

    def loop(
        self, body: ValueRef, inputs: dict[PortID, ValueRef], continue_port: PortID
    ) -> Callable[[PortID], ValueRef]:
        return self.add(Loop(body, inputs, continue_port))

    def map(
        self,
        body: ValueRef,
        inputs: dict[PortID, ValueRef],
    ) -> Callable[[PortID], ValueRef]:
        return self.add(Map(body, inputs))

    def if_else(self, pred: ValueRef, if_true: ValueRef, if_false: ValueRef):
        return self.add(IfElse(pred, if_true, if_false))

    def eager_if_else(self, pred: ValueRef, if_true: ValueRef, if_false: ValueRef):
        return self.add(EagerIfElse(pred, if_true, if_false))

    def output(self, inputs: dict[PortID, ValueRef]) -> None:
        self.add(Output(inputs))

    def add(self, node: NodeDef) -> Callable[[PortID], ValueRef]:
        idx = len(self.nodes)
        self.nodes.append(node)

        match node.type:
            case "output":
                if self.graph_output_idx is not None:
                    raise TierkreisError(
                        f"Graph already has output at index {self.graph_output_idx}"
                    )

                self.graph_output_idx = idx
            case "ifelse" | "eifelse":
                self.nodes[node.pred[0]].outputs.add(node.pred[1])
                self.nodes[node.if_true[0]].outputs.add(node.if_true[1])
                self.nodes[node.if_false[0]].outputs.add(node.if_false[1])
            case "input":
                self.graph_inputs.add(node.name)
            case "const" | "eval" | "function" | "loop" | "map":
                pass
            case _:
                assert_never(node)

        for i, port in node.inputs.values():
            self.nodes[i].outputs.add(port)

        return lambda k: (idx, k)

    def output_idx(self) -> NodeIndex:
        idx = self.graph_output_idx
        if idx is None:
            raise TierkreisError("Graph has no output index.")

        node = self.nodes[idx]
        if node.type != "output":
            raise TierkreisError(f"Expected output node at {idx} found {node}")

        return idx

    def remaining_inputs(self, provided_inputs: set[PortID]) -> set[PortID]:
        fixed_inputs = set(self.fixed_inputs.keys())
        if fixed_inputs & provided_inputs:
            raise TierkreisError(
                f"Fixed inputs {fixed_inputs}"
                f" should not intersect provided inputs {provided_inputs}."
            )

        actual_inputs = fixed_inputs.union(provided_inputs)
        return self.graph_inputs - actual_inputs


def graph_node_from_loc(
    node_location: Loc,
    graph: GraphData,
) -> tuple[NodeDef, GraphData]:
    """Assumes the first part of a loc can be found in current graph"""
    if len(graph.nodes) == 0:
        raise TierkreisError("Cannot convert location to node. Reason: Empty Graph")
    if node_location == "-":
        return Eval((-1, "body"), {}), graph

    step, remaining_location = node_location.pop_first()
    if isinstance(step, str):
        raise TierkreisError("Cannot convert location: Reason: Malformed Loc")
    (_, node_id) = step
    if node_id == -1:
        return Eval((-1, "body"), {}), graph
    node = graph.nodes[node_id]
    if remaining_location == Loc():
        return node, graph
    match node.type:
        case "eval":
            graph = _unwrap_graph(graph.nodes[node.graph[0]], node.type)
            node, graph = graph_node_from_loc(remaining_location, graph)
        case "loop" | "map":
            graph = _unwrap_graph(graph.nodes[node.body[0]], node.type)
            _, remaining_location = remaining_location.pop_first()  # Remove the M0/L0
            if len(remaining_location.steps()) < 2:
                return Eval((-1, "body"), node.inputs, node.outputs), graph

            node, graph = graph_node_from_loc(remaining_location, graph)
        case "const" | "function" | "input" | "output" | "ifelse" | "eifelse":
            pass
        case _:
            assert_never(node)

    return node, graph


def _unwrap_graph(node: NodeDef, node_type: str) -> GraphData:
    """Safely unwraps a const nodes GraphData."""
    if not isinstance(node, Const):
        raise TierkreisError(
            f"Cannot convert location to node. Reason: {node_type} does not wrap const"
        )
    match node.value:
        case GraphData() as graph:
            return graph
        case bytes() as thunk:
            return GraphData(**ptype_from_bytes(thunk, dict))
        case dict() as data:
            return GraphData(**data)

        case _:
            raise TierkreisError(
                "Cannot convert location to node. Reason: const value is not a graph"
            )



import json
import subprocess
from pathlib import Path

from tierkreis.controller.data.location import WorkerCallArgs
from tierkreis.exceptions import TierkreisError


class StdInOut:
    """Executes workers in an unix shell.

    Implements: :py:class:`tierkreis.controller.executor.protocol.ControllerExecutor`
    """

    def __init__(self, registry_path: Path, workflow_dir: Path) -> None:
        self.launchers_path = registry_path
        self.logs_path = workflow_dir / "logs"
        self.errors_path = workflow_dir / "logs"
        self.workflow_dir = workflow_dir

    def run(self, launcher_name: str, worker_call_args_path: Path) -> None:
        launcher_path = self.launchers_path / launcher_name
        self.errors_path = worker_call_args_path.parent / "errors"
        if not launcher_path.exists():
            raise TierkreisError(f"Launcher not found: {launcher_name}.")

        if launcher_path.is_dir() and not (launcher_path / "main.sh").exists():
            raise TierkreisError(f"Expected launcher file. Got {launcher_path}.")

        if launcher_path.is_dir() and not (launcher_path / "main.sh").is_file():
            raise TierkreisError(f"Expected launcher file. Got {launcher_path}/main.sh")

        if launcher_path.is_dir() and (launcher_path / "main.sh").is_file():
            launcher_path = launcher_path / "main.sh"

        with open(self.workflow_dir.parent / worker_call_args_path) as fh:
            call_args = WorkerCallArgs(**json.load(fh))

        input_file = self.workflow_dir.parent / list(call_args.inputs.values())[0]
        output_file = self.workflow_dir.parent / list(call_args.outputs.values())[0]
        done_path = self.workflow_dir.parent / call_args.done_path

        with open(self.workflow_dir.parent / self.logs_path, "a") as lfh:
            with open(self.workflow_dir.parent / self.errors_path, "a") as efh:
                proc = subprocess.Popen(
                    ["bash"],
                    start_new_session=True,
                    stdin=subprocess.PIPE,
                    stderr=efh,
                    stdout=lfh,
                )
                proc.communicate(
                    f"{launcher_path} <{input_file} >{output_file} && touch {done_path}".encode(),
                    timeout=10,
                )



from dataclasses import dataclass, field
from inspect import signature
from logging import getLogger
from pathlib import Path
import shutil
import subprocess
from typing import Callable, Self, get_origin
from tierkreis.codegen import format_method, format_model
from tierkreis.controller.data.models import PModel, is_portmapping
from tierkreis.controller.data.types import Struct, has_default, is_ptype
from tierkreis.exceptions import TierkreisError
from tierkreis.idl.spec import spec
from tierkreis.idl.models import GenericType, Interface, Method, Model, TypedArg

logger = getLogger(__name__)
WorkerFunction = Callable[..., PModel]


@dataclass
class Namespace:
    name: str
    methods: list[Method] = field(default_factory=lambda: [])
    models: set[Model] = field(default_factory=lambda: set())

    def _add_model_from_type(self, t: type) -> None:
        origin = get_origin(t)

        if is_ptype(t) and not isinstance(t, Struct) and not isinstance(origin, Struct):
            return

        if origin is None:
            origin = t

        annotations = origin.__annotations__
        portmapping_flag = True if is_portmapping(origin) else False
        decls = [TypedArg(k, GenericType.from_type(x)) for k, x in annotations.items()]
        model = Model(portmapping_flag, GenericType.from_type(t), decls)
        self.models.add(model)

    def add_function(self, func: WorkerFunction) -> None:
        name = func.__name__
        sig = signature(func)
        annotations = func.__annotations__
        generics: list[str] = [str(x) for x in func.__type_params__]
        in_annotations = {k: v for k, v in annotations.items() if k != "return"}
        ins = [
            TypedArg(k, GenericType.from_type(t.annotation), has_default(t))
            for k, t in sig.parameters.items()
        ]
        out = annotations["return"]

        for _, annotation in in_annotations.items():
            if not is_ptype(annotation):
                raise TierkreisError(f"Expected PType found {annotation} {annotations}")

        if not is_portmapping(out) and not is_ptype(out) and out is not None:
            raise TierkreisError(f"Expected PModel found {out}")

        method = Method(
            GenericType(name, generics),
            ins,
            GenericType.from_type(out),
            is_portmapping(out),
        )
        self.methods.append(method)
        [self._add_model_from_type(t) for t in annotations.values()]

    @classmethod
    def from_spec_file(cls, path: Path) -> "Namespace":
        with open(path) as fh:
            namespace_spec = spec(fh.read())
            return cls._from_spec(namespace_spec[0])

    @classmethod
    def _from_spec(cls, args: tuple[list[Model], Interface]) -> "Self":
        models = args[0]
        interface = args[1]
        namespace = cls(interface.name, models=set(models))
        for f in interface.methods:
            model = next((x for x in models if x.t == f.return_type), None)
            if model is not None:
                f.return_type_is_portmapping = model.is_portmapping
            namespace.methods.append(f)

        return namespace

    def stubs(self) -> str:
        functions = [format_method(self.name, f) for f in self.methods]
        functions_str = "\n\n".join(functions)

        models = sorted(list(self.models), key=lambda x: str(x.t.origin))
        models_str = "\n\n".join([format_model(x) for x in models])

        return f'''"""Code generated from {self.name} namespace. Please do not edit."""

from typing import Literal, NamedTuple, Sequence, TypeVar, Generic, Protocol, Union
from types import NoneType
from tierkreis.controller.data.models import TKR, OpaqueType
from tierkreis.controller.data.types import PType, Struct

{models_str}

{functions_str}
'''

    def write_stubs(self, stubs_path: Path) -> None:
        """Writes the type stubs to stubs_path.

        :param stubs_path: The location to write to.
        :type stubs_path: Path
        """
        with open(stubs_path, "w+") as fh:
            fh.write(self.stubs())

        ruff_binary = shutil.which("ruff")
        if ruff_binary:
            subprocess.run([ruff_binary, "format", stubs_path])
            subprocess.run([ruff_binary, "check", "--fix", stubs_path])
        else:
            logger.warning("No ruff binary found. Stubs will contain raw codegen.")



from pathlib import Path
from uuid import UUID
import pytest
from tierkreis.builder import GraphBuilder
from tierkreis.controller import run_graph
from tierkreis.controller.data.graph import GraphData
from tierkreis.controller.data.models import TKR
from tierkreis.controller.executor.hpc.job_spec import (
    JobSpec,
    MpiSpec,
    ResourceSpec,
)
from tierkreis.controller.executor.hpc.slurm import SLURMExecutor
from tierkreis.controller.storage.filestorage import ControllerFileStorage

from tests.executor.stubs import mpi_rank_info
from tierkreis.storage import read_outputs


def mpi_graph() -> GraphData:
    builder = GraphBuilder(outputs_type=TKR[str | None])
    mpi_result = builder.task(mpi_rank_info())
    builder.outputs(mpi_result)
    return builder.data


def job_spec() -> JobSpec:
    return JobSpec(
        job_name="test_job",
        account="test_usr",
        command="--allow-run-as-root /root/.local/bin/uv run /slurm_mpi_worker/main.py ",
        resource=ResourceSpec(nodes=2, memory_gb=None),
        walltime="00:15:00",
        mpi=MpiSpec(max_proc_per_node=1),
        extra_scheduler_args={"--open-mode=append": None},
        output_path=Path("./logs.log"),
        error_path=Path("./errors.log"),
    )


@pytest.mark.skip(reason="Needs SLURM setup.")
def test_slurm_with_mpi() -> None:
    g = mpi_graph()
    storage = ControllerFileStorage(
        UUID(int=22),
        name="mpi_graph",
        do_cleanup=True,
    )
    sbatch = str(
        Path(__file__).parent.parent.parent.parent / "infra/slurm_local/sbatch"
    )
    executor = SLURMExecutor(
        spec=job_spec(),
        registry_path=None,
        logs_path=storage.logs_path,
        command=sbatch,
    )
    run_graph(storage, executor, g, {})

    output = read_outputs(g, storage)

    assert output is not None
    assert output == "Rank 0 out of 2 on c1.\nRank 1 out of 2 on c2."



"""Core Tierkreis Label type."""


class Labels:
    """Special port labels used by builtin functions."""

    def __init__(self):
        raise RuntimeError("Do not instantiate")

    THUNK = "thunk"
    VALUE = "value"
    VARIANT_VALUE = "variant_value"
    CONTINUE = "continue"
    BREAK = "break"



import logging
import shutil
import subprocess
from pathlib import Path

from tierkreis.consts import TKR_DIR_KEY
from tierkreis.exceptions import TierkreisError

logger = logging.getLogger(__name__)


class UvExecutor:
    """Executes workers in an UV python environment.

    Implements: :py:class:`tierkreis.controller.executor.protocol.ControllerExecutor`
    """

    def __init__(
        self, registry_path: Path, logs_path: Path, env: dict[str, str] | None = None
    ) -> None:
        self.launchers_path = registry_path
        self.logs_path = logs_path
        self.errors_path = logs_path
        self.env = env or {}

    def run(
        self,
        launcher_name: str,
        worker_call_args_path: Path,
        uv_path: str | None = None,
    ) -> None:
        logging.basicConfig(
            format="%(asctime)s: %(message)s",
            datefmt="%Y-%m-%dT%H:%M:%S%z",
            filename=self.logs_path,
            filemode="a",
            level=logging.INFO,
        )
        self.errors_path = (
            self.logs_path.parent.parent / worker_call_args_path.parent / "errors"
        )
        logger.info("START %s %s", launcher_name, worker_call_args_path)

        if uv_path is None:
            uv_path = shutil.which("uv")
        if uv_path is None:
            raise TierkreisError("uv is required to use the uv_executor")

        worker_path = self.launchers_path / launcher_name

        env = self.env.copy()
        if "VIRTUAL_ENVIRONMENT" not in env:
            env["VIRTUAL_ENVIRONMENT"] = ""
        if TKR_DIR_KEY not in env:
            env[TKR_DIR_KEY] = str(self.logs_path.parent.parent)
        _error_path = self.errors_path.parent / "_error"

        with open(self.logs_path, "a") as lfh:
            with open(self.errors_path, "a") as efh:
                proc = subprocess.Popen(
                    ["bash"],
                    start_new_session=True,
                    stdin=subprocess.PIPE,
                    stderr=efh,
                    stdout=lfh,
                    cwd=worker_path,
                    env=env,
                )
                proc.communicate(
                    f"({uv_path} run main.py {worker_call_args_path} || touch {_error_path}) &".encode(),
                    timeout=10,
                )



"""Elementary parser and parser combinators.

Similar to https://github.com/drhagen/parsita
But https://github.com/drhagen/parsita/blob/83925f035d0777debfe5a6cb53b4944b4b5bcfe2/src/parsita/parsers/_sequential.py#L10
doesn't type check things correctly.
"""

import re
from typing import Callable, Never, overload

from tierkreis.exceptions import TierkreisError


class ParserError(TierkreisError): ...


class Parser[T]:
    fn: Callable[[str], tuple[T, str]]

    def __init__(self, fn: Callable[[str], tuple[T, str]]):
        self.fn = fn

    def __call__(self, ins: str) -> tuple[T, str]:
        ins = ins.strip()
        return self.fn(ins)

    def __or__[S](
        self, other: "Parser[S]" | Callable[[str], tuple[S, str]]
    ) -> "Parser[T|S]":
        """Try the left parser and only if it fails try the right parser."""

        def f(ins: str):
            try:
                return self(ins)
            except ParserError:
                return other(ins)

        return Parser(f)

    def __and__[S](
        self, other: "Parser[S]" | Callable[[str], tuple[S, str]]
    ) -> "Parser[tuple[T,S]]":
        """Use the left parser and then use the right parser on the remaining input."""

        def f(ins: str):
            s, remaining = self(ins)
            t, remaining = other(remaining)
            return (s, t), remaining

        return Parser(f)

    def __lshift__[S](
        self, other: "Parser[S]" | Callable[[str], tuple[S, str]]
    ) -> "Parser[T]":
        """Use the left parser and then the right parser but discard the result of the right parser."""

        def f(ins: str):
            t, remaining = self(ins)
            _, remaining = other(remaining)
            return t, remaining

        return Parser(f)

    def __rshift__[S](
        self, other: "Parser[S]" | Callable[[str], tuple[S, str]]
    ) -> "Parser[S]":
        """Use the left parser and then the right parser but discard the result of the left parser."""

        def f(ins: str):
            _, remaining = self(ins)
            s, remaining = other(remaining)
            return s, remaining

        return Parser(f)

    def opt(self) -> "Parser[T|None]":
        """Make the parser optional; if it fails then return None and carry on."""

        def f(ins: str):
            try:
                return self(ins)
            except ParserError:
                return None, ins

        return Parser(f)

    def map[A](self, fn: Callable[[T], A]) -> "Parser[A]":
        """Apply `fn` to transform the output of the parser."""

        def f(ins: str):
            t, remaining = self(ins)
            return fn(t), remaining

        return Parser(f)

    def coerce[A](self, a: A) -> "Parser[A]":
        """Shorthand for maps that don't need an argument.

        Not strictly speaking required."""

        def f(ins: str):
            t, remaining = self(ins)
            return a, remaining

        return Parser(f)

    def rep(self, sep: "Parser[str] | None" = None) -> "Parser[list[T]]":
        """Repeatedly apply a parser with an optional separator.

        The results of the separator parser are discarded."""

        def f(ins: str):
            outs: list[T] = []
            while True:
                try:
                    t, ins = self(ins)
                    if sep:
                        try:
                            _, ins = sep(ins)
                        except ParserError:
                            pass
                    outs.append(t)
                except ParserError:
                    break
            return outs, ins

        return Parser(f)

    def fail(self, entity: str) -> "Parser[Never]":
        """Fail early if we find something we don't support.

        Not strictly speaking required."""

        def f(ins: str):
            self(ins)
            raise TierkreisError(f"{entity} not supported.")

        return Parser(f)


@overload
def seq[A, B](*args: *tuple[Parser[A], Parser[B]]) -> Parser[tuple[A, B]]: ...
@overload
def seq[A, B, C](
    *args: *tuple[Parser[A], Parser[B], Parser[C]],
) -> Parser[tuple[A, B, C]]: ...
@overload
def seq[A, B, C, D](
    *args: *tuple[Parser[A], Parser[B], Parser[C], Parser[D]],
) -> Parser[tuple[A, B, C, D]]: ...
@overload
def seq[A, B, C, D, E](
    *args: *tuple[Parser[A], Parser[B], Parser[C], Parser[D], Parser[E]],
) -> Parser[tuple[A, B, C, D, E]]: ...
def seq(*args: Parser) -> Parser[tuple]:
    """Run a sequence of parsers one after the other
    and collect their outputs in a tuple."""

    def f(ins: str):
        outs = []
        for arg in args:
            s, ins = arg(ins)
            outs.append(s)
        return tuple(outs), ins

    return Parser(f)


def lit(*args: str) -> Parser[str]:
    """If the input starts with one of the strings in `args`
    then take the string off the stream and return it."""

    def f(ins: str):
        for a in args:
            if ins.startswith(a):
                return a, ins[len(a) :]

        raise ParserError(f"lit: expected {args} found '{ins[:20]}'")

    return Parser(f)


def reg(regex: str) -> Parser[str]:
    """If start of the input matches the `regex`
    then take the matching text off the stream and return it.

    Please don't pass match groups within the regex; they will be taken care of."""

    def f(ins: str):
        r = re.compile("^(" + regex + ")")

        if a := r.match(ins):
            return a.group(0), ins[a.end() :]

        raise ParserError(f"reg: expected regex {regex} found '{ins[:20]}'")

    return Parser(f)



import logging
from typing import assert_never

from tierkreis.controller.data.core import PortID, ValueRef
from tierkreis.controller.data.graph import (
    NodeDef,
)
from tierkreis.controller.data.location import Loc
from tierkreis.controller.storage.protocol import ControllerStorage

logger = logging.getLogger(__name__)


def in_edges(node: NodeDef) -> dict[PortID, ValueRef]:
    parents = {k: v for k, v in node.inputs.items()}

    match node.type:
        case "eval":
            parents["body"] = node.graph
        case "loop":
            parents["body"] = node.body
        case "map":
            parents["body"] = node.body
        case "ifelse":
            parents["pred"] = node.pred
        case "eifelse":
            parents["pred"] = node.pred
            parents["body_true"] = node.if_true
            parents["body_false"] = node.if_false
        case "const" | "function" | "input" | "output":
            pass
        case _:
            assert_never(node)

    return parents


def unfinished_inputs(
    storage: ControllerStorage, loc: Loc, node: NodeDef
) -> list[ValueRef]:
    ins = in_edges(node).values()
    ins = [x for x in ins if x[0] >= 0]  # inputs at -1 already finished
    return [x for x in ins if not storage.is_node_finished(loc.N(x[0]))]


def outputs_iter(storage: ControllerStorage, loc: Loc) -> list[tuple[int, PortID]]:
    eles = storage.read_output_ports(loc)
    return [(int(x.split("-")[-1]), x) for x in eles]



from pathlib import Path
from typing import Any, Protocol

from tierkreis.controller.data.core import PortID
from tierkreis.controller.data.graph import NodeDef
from tierkreis.controller.data.location import (
    Loc,
    OutputLoc,
    WorkerCallArgs,
)


class ControllerStorage(Protocol):
    """The storage protocol defines interaction with the *state* of the computation.

    The controller progresses the computation by committing the output values of a computation to the storage.
    They can then in turn be picked up as inputs of dependent calculations.
    For this purpose the protocol exposes *read_...* and *write_...* functions.
    """

    logs_path: Path

    def write_node_def(self, node_location: Loc, node: NodeDef) -> None:
        """Stores the definition of the node defined at the given location.

        Example::

            controller_storage.write_node_def(
                Loc(), Input(name="prediction", inputs={})
            )

        Will store the {"name": "pred", "inputs": {},"type": "input"} at location "-".

        :param node_location: The location of the node.
        :type node_location: Loc
        :param node: The definition of the node.
        :type node: NodeDef
        """
        ...

    def read_node_def(self, node_location: Loc) -> NodeDef:
        """Loads the definition of the node defined at the given location.

        Example::

            controller_storage.read_node_def(Loc())

        Will return the node definition stored at location "-".

        :param node_location: The location of the node.
        :type node_location: Loc
        :return: The definition of the node.
        :rtype: NodeDef
        """
        ...

    def write_worker_call_args(
        self,
        node_location: Loc,
        function_name: str,
        inputs: dict[PortID, OutputLoc],
        output_list: list[PortID],
    ) -> Path:
        """Stores the call arguments for the function node at the given location.

        Function nodes will always be invoked by a worker.
        For example a node `Func("builtins.iadd", {"a": a, "b": b})`
        would be stored as::

            {
                "function_name": "builtins.iadd",
                "inputs": {
                    "a": Loc("a"),
                    "b": Loc("b")
                },
                "outputs": {"value": Loc("value")}
            }

        where inputs and outputs link the ports of different nodes.

        :param node_location: The location of the node.
        :type node_location: Loc
        :param function_name: The name of the function being called.
        :type function_name: str
        :param inputs: The inputs to the function.
        :type inputs: dict[PortID, OutputLoc]
        :param output_list: The list of outputs from the function.
        :type output_list: list[PortID]
        :return: The path to the stored call arguments, such that the worker can find them.
        :rtype: Path
        """
        ...

    def read_worker_call_args(self, node_location: Loc) -> WorkerCallArgs:
        """Loads the call arguments for the function node at the given location.

        :param node_location: The location of the node.
        :type node_location: Loc
        :return: The call arguments for the worker.
        :rtype: WorkerCallArgs
        """
        ...

    def read_errors(self, node_location: Loc) -> str:
        """Loads the error logs for the node at the given location.

        :param node_location: The location of the node.
        :type node_location: Loc
        :return: The error logs for the node.
        :rtype: str
        """
        ...

    def node_has_error(self, node_location: Loc) -> bool:
        """Checks if the node at the given location has encountered an error.

        :param node_location: The location of the node.
        :type node_location: Loc
        :return: True if the node has error logs, False otherwise.
        :rtype: bool
        """
        ...

    def write_node_errors(self, node_location: Loc, error_logs: str) -> None:
        """Writes the error messages for the node at the given location.

        For function nodes this allows workers to forward error messages to the controller.

        :param node_location: The location of the node.
        :type node_location: Loc
        :param error_logs: The error logs to write.
        :type error_logs: str
        """
        ...

    def mark_node_finished(self, node_location: Loc) -> None:
        """Marks the node at the given location as finished.

        :param node_location: The location of the node.
        :type node_location: Loc
        """
        ...

    def is_node_finished(self, node_location: Loc) -> bool:
        """Checks if the node at the given location is finished.

        :param node_location: The location of the node.
        :type node_location: Loc
        :return: True if the node is finished, False otherwise.
        :rtype: bool
        """
        ...

    def link_outputs(
        self,
        new_location: Loc,
        new_port: PortID,
        old_location: Loc,
        old_port: PortID,
    ) -> None:
        """Connects the output port of one node to the output port of another.

        This is useful when unwrapping nodes such that the output of a nested node
        can be reused as the output of the parent node.

        :param new_location: The location of the new node.
        :type new_location: Loc
        :param new_port: The port of the new node.
        :type new_port: PortID
        :param old_location: The location of the old node.
        :type old_location: Loc
        :param old_port: The port of the old node.
        :type old_port: PortID
        """
        ...

    def write_output(
        self, node_location: Loc, output_name: PortID, value: bytes
    ) -> Path:
        """Writes the output value for port of the node at the given location.

        :param node_location: The location of the node.
        :type node_location: Loc
        :param output_name: The name of the output port.
        :type output_name: PortID
        :param value: The value to write to the output port.
        :type value: bytes
        :return: The path to the written output.
        :rtype: Path
        """
        ...

    def read_output(self, node_location: Loc, output_name: PortID) -> bytes:
        """Reads the output value for the port of the node at the given location.

        :param node_location: The location of the node.
        :type node_location: Loc
        :param output_name: The name of the output port.
        :type output_name: PortID
        :return: The output value for the port.
        :rtype: bytes
        """
        ...

    def read_output_ports(self, node_location: Loc) -> list[PortID]:
        """Checks which output ports are available for the node at the given location.

        :param node_location: The location of the node.
        :type node_location: Loc
        :return: The list of output ports for the node.
        :rtype: list[PortID]
        """
        ...

    def is_node_started(self, node_location: Loc) -> bool:
        """Checks if the node at the given location has started.

        :param node_location: The location of the node.
        :type node_location: Loc
        :return: True if the node has started, False otherwise.
        :rtype: bool
        """
        ...

    def read_metadata(self, node_location: Loc) -> dict[str, Any]:
        """Reads the metadata for the node at the given location.

        :param node_location: The location of the node.
        :type node_location: Loc
        :return: The metadata for the node.
        :rtype: dict[str, Any]
        """
        ...

    def write_metadata(self, node_location: Loc) -> None:
        """Writes the metadata for the node at the given location.

        :param node_location: The location of the node.
        :type node_location: Loc
        """
        ...

    def read_started_time(self, node_location: Loc) -> str | None:
        """Reads the start time of a node

        :param node_location: The location of the node
        :type node_location: Loc
        :return: A time string when node has started, else None.
        :rtype: str | None
        """
        ...

    def read_finished_time(self, node_location: Loc) -> str | None:
        """Reads the finish time of a node

        :param node_location: The location of the node
        :type node_location: Loc
        :return: A time string when node is completed, else None.
        :rtype: str | None
        """
        ...



from tierkreis.controller.executor.shell_executor import ShellExecutor
from tierkreis.controller.executor.uv_executor import UvExecutor
from tierkreis.controller.executor.multiple import MultipleExecutor
from tierkreis.controller.executor.hpc.pjsub import PJSUBExecutor


__all__ = ["ShellExecutor", "UvExecutor", "MultipleExecutor", "PJSUBExecutor"]



import pytest

from tierkreis.controller.data.location import Loc, NodeStep, get_last_index
from tierkreis.exceptions import TierkreisError

node_location_1 = Loc()
node_location_1 = node_location_1.N(1)
node_location_1 = node_location_1.L(0)
node_location_1 = node_location_1.N(3)
node_location_1 = node_location_1.L(2)
node_location_1 = node_location_1.N(0)
node_location_1 = node_location_1.M(7)
node_location_1 = node_location_1.N(10)


node_location_2 = Loc()
node_location_2 = node_location_2.N(0)
node_location_2 = node_location_2.L(0)
node_location_2 = node_location_2.N(3)
node_location_2 = node_location_2.N(8)
node_location_2 = node_location_2.N(0)

node_location_3 = Loc()
node_location_3 = node_location_3.N(0)

node_location_4 = Loc()


@pytest.mark.parametrize(
    ["node_location", "loc_str"],
    [
        (node_location_1, "-.N1.L0.N3.L2.N0.M7.N10"),
        (node_location_2, "-.N0.L0.N3.N8.N0"),
        (node_location_3, "-.N0"),
        (node_location_4, "-"),
    ],
)
def test_to_from_str(node_location: Loc, loc_str: str):
    node_location_str = str(node_location)
    assert node_location_str == loc_str

    new_loc = Loc(node_location_str)
    assert new_loc == node_location


@pytest.mark.parametrize(
    ["node_location", "node_step", "loc_str"],
    [
        (node_location_1, ("N", 1), "-.L0.N3.L2.N0.M7.N10"),
        (node_location_2, ("N", 0), "-.L0.N3.N8.N0"),
        (node_location_3, ("N", 0), "-"),
        (node_location_4, "-", ""),
    ],
)
def test_pop_first(node_location: Loc, node_step: NodeStep, loc_str: str) -> None:
    pop = node_location.pop_first()
    (step, remainder) = pop
    assert step == node_step
    assert remainder == Loc(loc_str)


@pytest.mark.parametrize(
    ["node_location", "node_step", "loc_str"],
    [
        (node_location_1, ("N", 10), "-.N1.L0.N3.L2.N0.M7"),
        (node_location_2, ("N", 0), "-.N0.L0.N3.N8"),
        (node_location_3, ("N", 0), "-"),
        (node_location_4, "-", ""),
    ],
)
def test_pop_last(node_location: Loc, node_step: NodeStep, loc_str: str) -> None:
    pop = node_location.pop_last()
    (step, remainder) = pop
    assert step == node_step
    assert remainder == Loc(loc_str)


def test_pop_empty() -> None:
    loc = Loc("")
    with pytest.raises(TierkreisError):
        loc.pop_first()
    with pytest.raises(TierkreisError):
        loc.pop_last()


def test_pop_first_multiple() -> None:
    loc = node_location_2
    pop = loc.pop_first()
    (step, remainder) = pop
    assert step == ("N", 0)
    assert remainder == Loc("-.L0.N3.N8.N0")
    pop = remainder.pop_first()
    (step, remainder) = pop
    assert step == ("L", 0)
    assert remainder == Loc("-.N3.N8.N0")
    pop = remainder.pop_first()
    (step, remainder) = pop
    assert step == ("N", 3)
    assert remainder == Loc("-.N8.N0")
    pop = remainder.pop_first()
    (step, remainder) = pop
    assert step == ("N", 8)
    assert remainder == Loc("-.N0")
    pop = remainder.pop_first()
    (step, remainder) = pop
    assert step == ("N", 0)
    assert remainder == Loc("-")
    pop = remainder.pop_first()
    (step, remainder) = pop
    assert step == "-"
    assert remainder == Loc("")


def test_pop_last_multiple() -> None:
    loc = node_location_2
    pop = loc.pop_last()
    (step, remainder) = pop
    assert step == ("N", 0)
    assert remainder == Loc("-.N0.L0.N3.N8")
    pop = remainder.pop_last()
    (step, remainder) = pop
    assert step == ("N", 8)
    assert remainder == Loc("-.N0.L0.N3")
    pop = remainder.pop_last()
    (step, remainder) = pop
    assert step == ("N", 3)
    assert remainder == Loc("-.N0.L0")
    pop = remainder.pop_last()
    (step, remainder) = pop
    assert step == ("L", 0)
    assert remainder == Loc("-.N0")
    pop = remainder.pop_last()
    (step, remainder) = pop
    assert step == ("N", 0)
    assert remainder == Loc("-")
    pop = remainder.pop_last()
    (step, remainder) = pop
    assert step == "-"
    assert remainder == Loc("")


@pytest.mark.parametrize(
    ["node_location", "index"],
    [
        (node_location_1, 10),
        (node_location_2, 0),
        (node_location_3, 0),
        (node_location_4, 0),
        (Loc().N(-1), -1),
    ],
)
def test_get_last_index(node_location: Loc, index: int) -> None:
    assert get_last_index(node_location) == index



from tierkreis import Labels
from tierkreis.controller.data.graph import GraphData


def doubler_plus() -> GraphData:
    g = GraphData()
    inp = g.input("doubler_input")
    intercept = g.input("intercept")
    two = g.const(2)
    mul = g.func("builtins.itimes", {"a": inp, "b": two})("value")
    out = g.func("builtins.iadd", {"a": mul, "b": intercept})("value")
    g.output({"doubler_output": out})
    return g


def simple_eval() -> GraphData:
    g = GraphData()
    zero = g.const(0)
    six = g.const(6)
    doubler_const = g.const(doubler_plus())
    e = g.eval(doubler_const, {"doubler_input": six, "intercept": zero})
    g.output({"simple_eval_output": e("doubler_output")})
    return g


def simple_partial() -> GraphData:
    g = GraphData()
    zero = g.const(0)
    six = g.const(6)
    doubler_const = g.const(doubler_plus())
    partial = g.eval(doubler_const, {"intercept": zero})("body")
    e = g.eval(partial, {"doubler_input": six})
    g.output({"simple_partial_output": e("doubler_output")})
    return g


def loop_body() -> GraphData:
    g = GraphData()
    a = g.input("loop_acc")
    one = g.const(1)
    N = g.const(10)

    a_plus = g.func("builtins.iadd", {"a": a, "b": one})("value")
    pred = g.func("builtins.igt", {"a": N, "b": a_plus})("value")
    g.output({"loop_acc": a_plus, "should_continue": pred})
    return g


def simple_loop() -> GraphData:
    g = GraphData()
    six = g.const(6)
    g_const = g.const(loop_body())
    loop = g.loop(g_const, {"loop_acc": six}, "should_continue")
    g.output({"simple_loop_output": loop("loop_acc")})
    return g


def simple_map() -> GraphData:
    g = GraphData()
    six = g.const(6)
    Ns_const = g.const(list(range(21)))
    Ns = g.func("builtins.unfold_values", {Labels.VALUE: Ns_const})
    doubler_const = g.const(doubler_plus())

    m = g.map(doubler_const, {"doubler_input": Ns("*"), "intercept": six})
    folded = g.func("builtins.fold_values", {"values_glob": m("*")})
    g.output({"simple_map_output": folded(Labels.VALUE)})
    return g


def simple_ifelse() -> GraphData:
    g = GraphData()
    pred = g.input("pred")
    one = g.const(1)
    two = g.const(2)
    out = g.if_else(pred, one, two)("value")
    g.output({"simple_ifelse_output": out})
    return g


def maps_in_series() -> GraphData:
    g = GraphData()
    zero = g.const(0)

    Ns_const = g.const(list(range(21)))
    Ns = g.func("builtins.unfold_values", {Labels.VALUE: Ns_const})
    doubler_const = g.const(doubler_plus())

    m = g.map(doubler_const, {"doubler_input": Ns("*"), "intercept": zero})

    m2 = g.map(doubler_const, {"doubler_input": m("*"), "intercept": zero})
    folded = g.func("builtins.fold_values", {"values_glob": m2("*")})
    g.output({"maps_in_series_output": folded(Labels.VALUE)})
    return g


def map_with_str_keys() -> GraphData:
    g = GraphData()
    zero = g.const(0)
    Ns_const = g.const({"one": 1, "two": 2, "three": 3})
    Ns = g.func("builtins.unfold_dict", {Labels.VALUE: Ns_const})
    doubler_const = g.const(doubler_plus())

    m = g.map(doubler_const, {"doubler_input": Ns("*"), "intercept": zero})
    folded = g.func("builtins.fold_dict", {"values_glob": m("*")})
    g.output({"map_with_str_keys_output": folded(Labels.VALUE)})
    return g


def factorial() -> GraphData:
    g = GraphData()
    minus_one = g.const(-1)
    one = g.const(1)
    n = g.input("n")
    fac = g.input("factorial")

    pred = g.func("builtins.igt", {"a": n, "b": one})("value")

    n_minus_one = g.func("builtins.iadd", {"a": minus_one, "b": n})("value")
    rec = g.eval(fac, {"n": n_minus_one, "factorial": fac})("factorial_output")
    if_true = g.func("builtins.itimes", {"a": n, "b": rec})("value")

    if_false = g.const(1)

    out = g.if_else(pred, if_true, if_false)("value")
    g.output({"factorial_output": out})
    return g


def simple_eagerifelse() -> GraphData:
    g = GraphData()
    pred = g.input("pred")
    one = g.const(1)
    two = g.const(2)
    out = g.eager_if_else(pred, one, two)(Labels.VALUE)
    g.output({"simple_eagerifelse_output": out})
    return g



from typing import NamedTuple, Protocol, SupportsIndex, runtime_checkable


PortID = str
NodeIndex = int
ValueRef = tuple[NodeIndex, PortID]


class EmptyModel(NamedTuple): ...


@runtime_checkable
class RestrictedNamedTuple[T](Protocol):
    """A NamedTuple whose members are restricted to being of type T."""

    def _asdict(self) -> dict[str, T]: ...
    def __getitem__(self, key: SupportsIndex, /) -> T: ...



# ruff: noqa: F821
from typing import Literal, NamedTuple
from tierkreis.builder import GraphBuilder
from tierkreis.controller.data.models import TKR, OpaqueType
from tierkreis.builtins.stubs import zip_impl, untuple
from tierkreis.aer_worker import (
    get_compiled_circuit as aer_compile,
    run_circuit as aer_run,
)
from tierkreis.qulacs_worker import (
    get_compiled_circuit as qulacs_compile,
    run_circuit as qulacs_run,
)
from tierkreis.builtins.stubs import str_eq

type BackendResult = OpaqueType["pytket.backends.backendresult.BackendResult"]
type Circuit = OpaqueType["pytket._tket.circuit.Circuit"]


class SimulateJobInputs(NamedTuple):
    simulator_name: TKR[Literal["aer", "qulacs"]]
    circuits: TKR[list[Circuit]]
    n_shots: TKR[list[int]]
    compilation_optimisation_level: TKR[int]


class SimulateJobInputsSingle(NamedTuple):
    simulator_name: TKR[Literal["aer", "qulacs"]]
    circuit_shots: TKR[tuple[Circuit, int]]
    compilation_optimisation_level: TKR[int]


def aer_simulate_single():
    g = GraphBuilder(SimulateJobInputsSingle, TKR[BackendResult])
    circuit_shots = g.task(untuple(g.inputs.circuit_shots))

    compiled_circuit = g.task(
        aer_compile(
            circuit=circuit_shots.a,
            optimisation_level=g.inputs.compilation_optimisation_level,
        )
    )
    res = g.task(aer_run(compiled_circuit, circuit_shots.b))
    g.outputs(res)
    return g


def qulacs_simulate_single():
    g = GraphBuilder(SimulateJobInputsSingle, TKR[BackendResult])
    circuit_shots = g.task(untuple(g.inputs.circuit_shots))

    compiled_circuit = g.task(
        qulacs_compile(
            circuit=circuit_shots.a,
            optimisation_level=g.inputs.compilation_optimisation_level,
        )
    )
    res = g.task(qulacs_run(compiled_circuit, circuit_shots.b))
    g.outputs(res)
    return g


def compile_simulate_single():
    g = GraphBuilder(SimulateJobInputsSingle, TKR[BackendResult])

    aer_res = g.eval(aer_simulate_single(), g.inputs)
    qulacs_res = g.eval(qulacs_simulate_single(), g.inputs)
    res = g.ifelse(
        g.task(str_eq(g.inputs.simulator_name, g.const("aer"))), aer_res, qulacs_res
    )

    g.outputs(res)
    return g


def compile_simulate():
    g = GraphBuilder(SimulateJobInputs, TKR[list[BackendResult]])

    circuits_shots = g.task(zip_impl(g.inputs.circuits, g.inputs.n_shots))

    inputs = g.map(
        lambda x: SimulateJobInputsSingle(
            simulator_name=g.inputs.simulator_name,
            circuit_shots=x,
            compilation_optimisation_level=g.inputs.compilation_optimisation_level,
        ),
        circuits_shots,
    )
    res = g.map(compile_simulate_single(), inputs)

    g.outputs(res)
    return g



import json
import pytest
import sys
from pathlib import Path
from unittest import mock
from uuid import UUID

from tierkreis.cli.tkr import load_graph, _load_inputs, main
from tierkreis.controller.data.graph import GraphData
from tierkreis.controller.data.types import PType
from tierkreis.exceptions import TierkreisError

from tests.controller.sample_graphdata import simple_eval

simple_eval_graph = simple_eval()

graph_params = [
    ("tests.controller.sample_graphdata:simple_eval", simple_eval_graph),
    ("tierkreis/tests/controller/sample_graphdata.py:simple_eval", simple_eval_graph),
]


@pytest.mark.parametrize("input,graph", graph_params, ids=["load_module", "load_file"])
def test_load_graph(input: str, graph: GraphData) -> None:
    assert load_graph(input) == graph


def test_load_graph_invalid() -> None:
    with pytest.raises(FileNotFoundError):
        load_graph("sample_graphdata.py:simple_eval")
    with pytest.raises(ModuleNotFoundError):
        load_graph("sample_graphdata:simple_eval")
    with pytest.raises(TierkreisError):
        load_graph("invalid_arg")
    with pytest.raises(ModuleNotFoundError):
        load_graph("at_least_a:colon")


input_params = [
    (
        ["tierkreis/tests/cli/data/data.json"],
        {
            "a_string": b'"string"',
            "a_value": b"5",
        },
    ),
    (
        [
            "input1:tierkreis/tests/cli/data/input1",
            "input2:tierkreis/tests/cli/data/input2",
        ],
        {"input1": b"test\n", "input2": {"a": 5, "b": "string"}},
    ),
]


@pytest.mark.parametrize(
    "input,result", input_params, ids=["json_input", "binary_input"]
)
def test_load_inputs(input: list[str], result: dict[str, PType]) -> None:
    assert _load_inputs(input) == result


def test_load_inputs_invalid() -> None:
    with pytest.raises(FileNotFoundError):
        _load_inputs(["data.json"])
    with pytest.raises(FileNotFoundError):
        _load_inputs(["at_least_a:colon"])
    with pytest.raises(TierkreisError):
        _load_inputs(["wrong_format"])


default_args = [
    "tkr",
    "run",
    "--run-id",
    "1860",
    "-v",
    "-o",
    "-n",
    "500",
    "-p",
    "0.02",
    "-r",
    "--name",
    "test_name",
    "--uv",
    "--registry-path",
    "tests/controller/sample_graphdata",
]

cli_params = [
    (
        default_args + ["-f", "tierkreis/tests/cli/data/sample_graph"],
        {"simple_eval_output": 12},
    ),
    (
        default_args
        + [
            "-g",
            "tests.controller.sample_graphdata:factorial",
            "-i",
            "n:tierkreis/tests/cli/data/n",
            "factorial:tierkreis/tests/cli/data/factorial",
        ],
        {"factorial_output": 120},
    ),
]


@pytest.mark.parametrize(
    "args,result", cli_params, ids=["simple_eval_cli", "factorial_cli"]
)
def test_end_to_end(args: list[str], result: dict[str, bytes]) -> None:
    with mock.patch.object(sys, "argv", args):
        main()
    for key, value in result.items():
        with open(
            Path.home()
            / ".tierkreis"
            / "checkpoints"
            / str(UUID(int=1860))
            / f"-/outputs/{key}",
            "rb",
        ) as fh:
            c = json.loads(fh.read())
        assert c == value



from types import NoneType
import pytest
from tierkreis.codegen import format_generic_type
from tierkreis.controller.data.types import PType
from tierkreis.idl.models import GenericType

formats: list[tuple[type[PType], str]] = [
    (bool, "bool"),
    (int, "int"),
    (float, "float"),
    (str, "str"),
    (bytes, "bytes"),
    (NoneType, "NoneType"),
    (list[str], "list[str]"),
    (list[str | list[str | int]], "list[Union[str, list[Union[str, int]]]]"),
    (tuple[str], "tuple[str]"),
    (
        tuple[str | list[str | int], NoneType],
        "tuple[Union[str, list[Union[str, int]]], NoneType]",
    ),
]


@pytest.mark.parametrize("ttype,expected", formats)
def test_format_ttype(ttype: type[PType], expected: str):
    generic_type = GenericType.from_type(ttype)

    assert format_generic_type(generic_type, False, False) == expected



import pytest
from tierkreis.exceptions import TierkreisError
from tierkreis.controller.data.graph import GraphData


def test_only_one_output():
    with pytest.raises(TierkreisError):
        g = GraphData()
        g.output({"one": g.const(1)})
        g.output({"two": g.const(2)})



from pathlib import Path
from typing import Callable
from tierkreis.controller.executor.hpc.hpc_executor import run_hpc_executor
from tierkreis.controller.executor.hpc.job_spec import JobSpec


_COMMAND_PREFIX = "#SBATCH"


def generate_slurm_script(spec: JobSpec) -> str:
    # 1. Shebang and file header
    lines = [
        """#!/bin/bash
#
# SLURM Job Script generated by TIERKREIS

# --- Core Job Specifications ---"""
    ]
    # 2. Name
    lines.append(f"{_COMMAND_PREFIX} --job-name={spec.job_name}")
    # 3. Resources (node exclusive)
    lines.append(f"{_COMMAND_PREFIX} --time={spec.walltime}")
    lines.append(f"{_COMMAND_PREFIX} --nodes={spec.resource.nodes}")
    if spec.resource.cores_per_node is not None:
        lines.append(
            f"{_COMMAND_PREFIX} --cpus-per-task={spec.resource.cores_per_node}"
        )
    if spec.resource.memory_gb is not None:
        lines.append(f"{_COMMAND_PREFIX} --mem={spec.resource.memory_gb}G")
    if spec.resource.gpus_per_node is not None:
        lines.append(f"{_COMMAND_PREFIX} --gpus-per-node={spec.resource.gpus_per_node}")
    if spec.queue is not None:
        lines.append(f"{_COMMAND_PREFIX} --partition={spec.queue}")

    # 4. User settings
    lines.append("\n# --- User Details ---")
    if spec.account is not None:
        lines.append(f"{_COMMAND_PREFIX} --account={spec.account}")
    if spec.user is not None:
        if spec.user.mail is not None:
            lines.append(f"{_COMMAND_PREFIX} --mail-type=END")  # end only
            lines.append(f"{_COMMAND_PREFIX} --mail-user={spec.user.mail}")

    # 5. Output and Error handling
    lines.append("\n# --- Output and Error Handling ---")
    if spec.error_path is not None:
        lines.append(f"{_COMMAND_PREFIX} --error={spec.error_path}")
    if spec.output_path is not None:
        lines.append(f"{_COMMAND_PREFIX} --output={spec.output_path}")

    # 6. MPI, #TODO check if this makes sense
    if spec.mpi is not None:
        lines.append("\n# --- MPI ---")
        if spec.mpi.proc is not None:
            lines.append(f"{_COMMAND_PREFIX} --ntasks={spec.mpi.proc}")
        if spec.mpi.max_proc_per_node is not None:
            lines.append(
                f"{_COMMAND_PREFIX} --ntasks-per-node={spec.mpi.max_proc_per_node}"
            )

    # 7. User specific
    lines.append("\n# --- User-Specific Arguments ---")
    for key, value in spec.extra_scheduler_args.items():
        lines.append(f"{_COMMAND_PREFIX} {key} {value if value is not None else ''}")

    # 8. Environment
    lines.append("\n# --- Environment Setup ---")
    if spec.environment != {}:
        env = ",".join(
            f"{key}={value if value else '""'}"
            for key, value in spec.environment.items()
        )
        lines.append(f"--export={env}")
    # 9. Container logic

    # 10. User Command, (prologue), command, (epilogue)
    lines.append("\n# --- User Command ---")
    if spec.mpi is not None:
        if spec.mpi.max_proc_per_node is None:
            spec.mpi.max_proc_per_node = 1
        lines.append(
            f"mpirun -n {spec.resource.nodes * spec.mpi.max_proc_per_node} {spec.command}"
        )
    else:
        lines.append(spec.command)

    return "\n".join(lines)


class SLURMExecutor:
    def __init__(
        self,
        registry_path: Path | None,
        logs_path: Path,
        spec: JobSpec,
        command: str = "sbatch",
    ) -> None:
        self.launchers_path = registry_path
        self.logs_path = logs_path
        self.errors_path = logs_path
        self.spec = spec
        self.script_fn: Callable[[JobSpec], str] = generate_slurm_script
        self.command = command

    def run(self, launcher_name: str, worker_call_args_path: Path) -> None:
        self.errors_path = (
            self.logs_path.parent.parent / worker_call_args_path.parent / "errors"
        )
        run_hpc_executor(self, launcher_name, worker_call_args_path)



import json
from glob import glob
import os
from pathlib import Path

from tierkreis.consts import TKR_DIR_KEY
from tierkreis.controller.data.location import WorkerCallArgs


class WorkerFileStorage:
    def __init__(self, tierkreis_dir: Path | None = None) -> None:
        if tierkreis_dir is not None:
            self.tierkreis_dir = tierkreis_dir
        elif tierkreis_dir_str := os.environ.get(TKR_DIR_KEY):
            self.tierkreis_dir = Path(tierkreis_dir_str).resolve()
        else:
            self.tierkreis_dir = Path.home() / ".tierkreis" / "checkpoints"

    def resolve(self, path: Path | str) -> Path:
        path = Path(path)
        return path if path.is_absolute() else self.tierkreis_dir / path

    def read_call_args(self, path: Path) -> WorkerCallArgs:
        with open(self.resolve(path), "r") as fh:
            return WorkerCallArgs(**json.loads(fh.read()))

    def read_input(self, path: Path) -> bytes:
        with open(self.resolve(path), "rb") as fh:
            return fh.read()

    def write_output(self, path: Path, value: bytes) -> None:
        with open(self.resolve(path), "wb+") as fh:
            fh.write(value)

    def glob(self, path_string: str) -> list[str]:
        return glob(str(self.resolve(path_string)))

    def mark_done(self, path: Path) -> None:
        self.resolve(path).touch()

    def write_error(self, path: Path, error_logs: str) -> None:
        with open(self.resolve(path), "w+") as f:
            f.write(error_logs)



import logging
import subprocess
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Callable, Protocol

from tierkreis.consts import TKR_DIR_KEY
from tierkreis.controller.executor.hpc.job_spec import JobSpec
from tierkreis.exceptions import TierkreisError

logger = logging.getLogger(__name__)


class HPCExecutor(Protocol):
    launchers_path: Path | None
    logs_path: Path
    errors_path: Path
    spec: JobSpec
    script_fn: Callable[[JobSpec], str]
    command: str


def generate_script(
    template_fn: Callable[[JobSpec], str], spec: JobSpec, path: Path
) -> None:
    with open(path, "w+", encoding="utf-8") as fh:
        fh.write(template_fn(spec))


def run_hpc_executor(
    executor: HPCExecutor, launcher_name: str, worker_call_args_path: Path
) -> None:
    logging.basicConfig(
        format="%(asctime)s: %(message)s",
        datefmt="%Y-%m-%dT%H:%M:%S%z",
        filename=executor.logs_path,
        filemode="a",
        level=logging.INFO,
        force=True,
    )
    logger.info("START %s %s", launcher_name, worker_call_args_path)

    spec = executor.spec.model_copy()
    if executor.launchers_path:
        spec.command = f"cd {executor.launchers_path}/{launcher_name} && {spec.command}"

    spec.command += " " + str(worker_call_args_path)
    submission_cmd = [executor.command]
    if spec.output_path is None:
        submission_cmd += ["-o", str(executor.logs_path)]
    else:
        submission_cmd += ["-o", str(spec.output_path)]
    if spec.error_path is None:
        submission_cmd += ["-e", str(executor.errors_path)]
    else:
        submission_cmd += ["-e", str(spec.error_path)]
    if spec.include_no_check_directory_flag:
        submission_cmd += ["--no-check-directory"]

    if TKR_DIR_KEY not in spec.environment:  # User can override by setting TKR_DIR
        spec.environment[TKR_DIR_KEY] = str(executor.logs_path.parent.parent)

    with NamedTemporaryFile(
        mode="w+",
        delete=True,
        suffix=".sh",
        prefix=f"{spec.job_name}-",
    ) as script_file:
        generate_script(executor.script_fn, spec, Path(script_file.name))
        submission_cmd.append(script_file.name)

        process = subprocess.run(
            submission_cmd,
            start_new_session=True,
            capture_output=True,
            universal_newlines=True,
        )
    if process.returncode != 0:
        with open(executor.errors_path, "a") as efh:
            efh.write("Error from script")
            efh.write(process.stderr)
        raise TierkreisError(f"Executor failed with return code {process.returncode}")
    logger.info("Submitted job with return code %s", process.stdout.rstrip())



from inspect import isclass
from pydantic import BaseModel
from tierkreis.controller.data.types import DictConvertible, ListConvertible, _is_union
from tierkreis.idl.models import GenericType, Method, Model, TypedArg

NO_QA_STR = " # noqa: F821 # fmt: skip"


def format_ptype(ptype: type | str) -> str:
    if isinstance(ptype, str):
        return ptype

    if isclass(ptype) and issubclass(
        ptype, (DictConvertible, ListConvertible, BaseModel)
    ):
        return f'OpaqueType["{ptype.__module__}.{ptype.__qualname__}"]'

    if _is_union(ptype):
        return "Union"

    return ptype.__qualname__


def format_generic_type(
    generictype: GenericType | str, include_bound: bool, is_tkr: bool
) -> str:
    bound_str = ": PType" if include_bound else ""
    if isinstance(generictype, str):
        out = generictype + bound_str
        return f"TKR[{out}]" if is_tkr else out

    origin_str = format_ptype(generictype.origin)

    generics = [format_generic_type(x, include_bound, False) for x in generictype.args]
    generics_str = f"[{', '.join(generics)}]" if generictype.args else ""

    out = f"{origin_str}{generics_str}"
    return f"TKR[{out}]" if is_tkr else out


def format_typed_arg(typed_arg: TypedArg, is_portmaping: bool) -> str:
    type_str = format_generic_type(typed_arg.t, False, not is_portmaping)
    default_str = " | None = None " if typed_arg.has_default else ""
    return f"{typed_arg.name}: {type_str}{default_str} {NO_QA_STR}"


def format_model(model: Model) -> str:
    is_portmapping = model.is_portmapping
    outs = [format_typed_arg(x, not is_portmapping) for x in model.decls]
    outs.sort()
    outs_str = "\n    ".join(outs)

    bases = ["NamedTuple"] if is_portmapping else ["Struct", "Protocol"]

    return f"""
class {format_generic_type(model.t, True, False)}({", ".join(bases)}):
    {outs_str}
"""


def format_method(namespace_name: str, fn: Method) -> str:
    ins = [format_typed_arg(x, False) for x in fn.args]
    ins_str = "\n    ".join(ins)
    class_name = format_generic_type(
        fn.return_type, False, not fn.return_type_is_portmapping
    )

    bases = ["NamedTuple"]

    return f"""class {format_generic_type(fn.name, True, False)}({", ".join(bases)}):
    {ins_str}

    @staticmethod
    def out() -> type[{class_name}]:{NO_QA_STR}
        return {class_name}{NO_QA_STR}

    @property
    def namespace(self) -> str:
        return "{namespace_name}" """



from pathlib import Path
from typing import Protocol


class ControllerExecutor(Protocol):
    """The executor protocol defines how to run workers.

    An executor is responsible for running a worker(binary) in a given environment.

    """

    def run(self, launcher_name: str, worker_call_args_path: Path) -> None:
        """Run the node defined by the node_definition path.

        Specifies the worker to run by its launcher name.
        For example the function "builtins.iadd" will call the builtins worker's iadd function.
        The call arguments for the function call are retrieved retrieved from its location.

        :param launcher_name: module description of launcher to run.
        :type launcher_name: str
        :param worker_call_args_path: Location of the worker call args.
        :type worker_call_args_path: Path
        """

    ...



import logging
from base64 import b64decode, b64encode
import collections.abc
from inspect import Parameter, _empty, isclass
from itertools import chain
import json
from types import NoneType, UnionType
from typing import (
    Annotated,
    Any,
    Mapping,
    Protocol,
    Self,
    Sequence,
    TypeVar,
    Union,
    assert_never,
    cast,
    get_args,
    get_origin,
    runtime_checkable,
)

from pydantic import BaseModel, ValidationError
from pydantic._internal._generics import get_args as pydantic_get_args
from tierkreis.controller.data.core import RestrictedNamedTuple
from tierkreis.exceptions import TierkreisError
from typing_extensions import TypeIs


@runtime_checkable
class DictConvertible(Protocol):
    def to_dict(self) -> dict: ...
    @classmethod
    def from_dict(cls, arg: dict, /) -> "Self": ...


@runtime_checkable
class ListConvertible(Protocol):
    def to_list(self) -> list: ...
    @classmethod
    def from_list(cls, arg: list, /) -> "Self": ...


type Container[T] = (
    T
    | list[Container[T]]
    | Sequence[Container[T]]
    | tuple[Container[T], ...]
    | dict[str, Container[T]]
    | Mapping[str, Container[T]]
)
type ElementaryType = (
    bool
    | int
    | float
    | str
    | NoneType
    | bytes
    | DictConvertible
    | ListConvertible
    | BaseModel
)
type JsonType = Container[ElementaryType]
logger = logging.getLogger(__name__)


@runtime_checkable
class Struct(RestrictedNamedTuple[JsonType], Protocol): ...


_StructPType = JsonType | Struct
PType = Container[_StructPType]
"""A restricted subset of Python types that can be used to annotate
worker functions for automatic codegen of graph builder stubs."""


class TierkreisEncoder(json.JSONEncoder):
    """Encode bytes also."""

    def default(self, o):
        if isinstance(o, bytes):
            return {"__tkr_bytes__": True, "bytes": b64encode(o).decode()}

        return super().default(o)


class TierkreisDecoder(json.JSONDecoder):
    """Decode bytes also."""

    def __init__(self, **kwargs):
        kwargs.setdefault("object_hook", self._object_hook)
        super().__init__(**kwargs)

    def _object_hook(self, d):
        """Try to decode an object containing bytes."""
        if "__tkr_bytes__" in d and "bytes" in d:
            return b64decode(d["bytes"])
        return d


def _is_union(o: object) -> bool:
    return (
        get_origin(o) == UnionType
        or get_origin(o) == Union
        or o == Union
        or o == UnionType
    )


def _is_generic(o) -> TypeIs[type[TypeVar]]:
    return isinstance(o, TypeVar)


def _is_list(ptype: object) -> TypeIs[type[Sequence[PType]]]:
    return get_origin(ptype) == collections.abc.Sequence or get_origin(ptype) is list


def _is_mapping(ptype: object) -> TypeIs[type[Mapping[str, PType]]]:
    return get_origin(ptype) is collections.abc.Mapping or get_origin(ptype) is dict


def _is_tuple(o: object) -> TypeIs[type[tuple[Any, ...]]]:
    return get_origin(o) is tuple


def is_ptype(annotation: Any) -> TypeIs[type[PType]]:
    if get_origin(annotation) is Annotated:
        return is_ptype(get_args(annotation)[0])

    if _is_generic(annotation):
        return True

    if (
        _is_union(annotation)
        or _is_tuple(annotation)
        or _is_list(annotation)
        or _is_mapping(annotation)
    ):
        return all(is_ptype(x) for x in get_args(annotation))

    elif isclass(annotation) and issubclass(
        annotation, (DictConvertible, ListConvertible, BaseModel, Struct)
    ):
        return True

    elif annotation in get_args(ElementaryType.__value__):
        return True

    origin = get_origin(annotation)
    if origin is not None:
        return is_ptype(origin) and all(is_ptype(x) for x in get_args(annotation))

    else:
        return False


def ser_from_ptype(ptype: PType) -> Any | bytes:
    match ptype:
        case bytes() | bytearray() | memoryview():
            # Top level bytes should be a clean pass-through.
            return bytes(ptype)
        case bool() | int() | float() | str() | NoneType() | TypeVar():
            return ptype
        case Struct():
            return {k: ser_from_ptype(p) for k, p in ptype._asdict().items()}
        case collections.abc.Sequence():
            return [ser_from_ptype(p) for p in ptype]
        case collections.abc.Mapping():
            return {k: ser_from_ptype(p) for k, p in ptype.items()}
        case DictConvertible():
            return ser_from_ptype(ptype.to_dict())
        case ListConvertible():
            return ser_from_ptype(ptype.to_list())
        case BaseModel():
            return ptype.model_dump(mode="json")
        case _:
            assert_never(ptype)


def bytes_from_ptype(ptype: PType) -> bytes:
    ser = ser_from_ptype(ptype)
    match ptype:
        case bytes():
            # Top level bytes should be a clean pass-through.
            return ptype
        case _:
            return json.dumps(ser, cls=TierkreisEncoder).encode()


def coerce_from_annotation[T: PType](ser: Any, annotation: type[T]) -> T:
    if get_origin(annotation) is Annotated:
        return coerce_from_annotation(ser, get_args(annotation)[0])

    if _is_union(annotation):
        for t in get_args(annotation):
            try:
                return coerce_from_annotation(ser, t)
            except (AssertionError, ValidationError):
                logger.debug(f"Tried deserialising as {t}")
        raise TierkreisError(f"Could not deserialise {ser} as {annotation}")

    origin = get_origin(annotation)
    if origin is None:
        origin = annotation

    if ser is None:
        return ser

    if isinstance(origin, TypeVar):
        # Required to support generic parameters in functions,
        # we can't really make a judgement about what type it
        # should be deserialised in this case and so have to
        # just return the value in its "raw" form.
        return ser

    if issubclass(origin, (bool, int, float, str, bytes, NoneType)):
        return ser

    if issubclass(origin, DictConvertible):
        assert issubclass(annotation, origin)
        return annotation.from_dict(ser)

    if issubclass(origin, ListConvertible):
        assert issubclass(annotation, origin)
        return annotation.from_list(ser)

    if issubclass(origin, BaseModel):
        assert issubclass(annotation, origin)
        return annotation(**ser)

    if issubclass(origin, Struct):
        return cast(T, origin(**ser))

    if issubclass(origin, collections.abc.Sequence):
        args = get_args(annotation)
        if len(args) == 0:
            return ser

        return cast(T, [coerce_from_annotation(x, args[0]) for x in ser])

    if issubclass(origin, collections.abc.Mapping):
        args = get_args(annotation)
        if len(args) == 0:
            return ser

        return cast(T, {k: coerce_from_annotation(v, args[1]) for k, v in ser.items()})

    assert_never(ser)


def ptype_from_bytes[T: PType](bs: bytes, annotation: type[T] | None = None) -> T:
    if isclass(annotation) and issubclass(annotation, bytes):
        return cast(T, bs)

    try:
        j = json.loads(bs, cls=TierkreisDecoder)
    except json.JSONDecodeError as err:
        if annotation is None:
            return cast(T, bs)
        raise err

    if annotation is None:
        return j

    return coerce_from_annotation(j, annotation)


def generics_in_ptype(ptype: type[PType]) -> set[str]:
    if _is_generic(ptype):
        return {str(ptype)}

    if _is_union(ptype) or _is_tuple(ptype) or _is_list(ptype) or _is_mapping(ptype):
        return set(chain(*[generics_in_ptype(x) for x in get_args(ptype)]))

    origin = get_origin(ptype)
    if origin is not None:
        return generics_in_ptype(origin)

    if issubclass(ptype, (bool, int, float, str, bytes, NoneType)):
        return set()

    if issubclass(ptype, (DictConvertible, ListConvertible, Struct)):
        return set()

    if issubclass(ptype, BaseModel):
        return set((str(x) for x in pydantic_get_args(ptype)))

    assert_never(ptype)


def has_default(t: Parameter) -> bool:
    return not (isclass(t.default) and issubclass(t.default, _empty))



import json
import logging
import os
import shutil
from datetime import datetime
from pathlib import Path
from time import time_ns
from typing import Any
from uuid import UUID

from tierkreis.controller.data.core import PortID
from tierkreis.controller.data.graph import NodeDef, NodeDefModel
from tierkreis.controller.data.location import (
    Loc,
    OutputLoc,
    WorkerCallArgs,
)
from tierkreis.exceptions import TierkreisError

logger = logging.getLogger(__name__)


class ControllerFileStorage:
    """Default file-base storage for the controller.

    All information is stored in the file system, by default in `~/.tierkreis/checkpoints`.
    Uses file systems links to map outputs to new nodes.
    Invalid read operations raise a FileNotFoundError.

    Implements: :py:class:`tierkreis.controller.storage.protocol.ControllerStorage`
    """

    def __init__(
        self,
        workflow_id: UUID,
        name: str | None = None,
        tierkreis_directory: Path = Path.home() / ".tierkreis" / "checkpoints",
        do_cleanup: bool = False,
    ) -> None:
        self.workflow_id = workflow_id
        self.tkr_dir = tierkreis_directory
        self.workflow_dir: Path = tierkreis_directory / str(workflow_id)
        self.workflow_dir.mkdir(parents=True, exist_ok=True)
        self.logs_path = self.workflow_dir / "logs"
        self.name = name
        if do_cleanup:
            self.clean_graph_files()

    def _nodedef_path(self, node_location: Loc) -> Path:
        path = self.workflow_dir / str(node_location) / "nodedef"
        path.parent.mkdir(parents=True, exist_ok=True)
        return path

    def _worker_call_args_path(self, node_location: Loc) -> Path:
        path = self.workflow_dir / str(node_location) / "definition"
        path.parent.mkdir(parents=True, exist_ok=True)
        return path

    def _metadata_path(self, node_location: Loc) -> Path:
        path = self.workflow_dir / str(node_location) / "_metadata"
        path.parent.mkdir(parents=True, exist_ok=True)
        return path

    def _outputs_dir(self, node_location: Loc) -> Path:
        path = self.workflow_dir / str(node_location) / "outputs"
        path.mkdir(parents=True, exist_ok=True)
        return path

    def _output_path(self, node_location: Loc, port_name: PortID) -> Path:
        path = self._outputs_dir(node_location) / port_name
        path.parent.mkdir(parents=True, exist_ok=True)
        return path

    def _done_path(self, node_location: Loc) -> Path:
        path = self.workflow_dir / str(node_location) / "_done"
        path.parent.mkdir(parents=True, exist_ok=True)
        return path

    def _error_path(self, node_location: Loc) -> Path:
        path = self.workflow_dir / str(node_location) / "_error"
        path.parent.mkdir(parents=True, exist_ok=True)
        return path

    def _error_logs_path(self, node_location: Loc) -> Path:
        path = self.workflow_dir / str(node_location) / "errors"
        path.parent.mkdir(parents=True, exist_ok=True)
        return path

    def clean_graph_files(self) -> None:
        uid = os.getuid()
        tmp_dir = Path(f"/tmp/{uid}/tierkreis/archive/{self.workflow_id}/{time_ns()}")
        tmp_dir.mkdir(parents=True)
        if self.workflow_dir.exists():
            shutil.move(self.workflow_dir, tmp_dir)

    def write_node_def(self, node_location: Loc, node: NodeDef):
        with open(self._nodedef_path(node_location), "w+") as fh:
            fh.write(NodeDefModel(root=node).model_dump_json())

    def read_node_def(self, node_location: Loc) -> NodeDef:
        with open(self._nodedef_path(node_location)) as fh:
            return NodeDefModel(**json.load(fh)).root

    def write_worker_call_args(
        self,
        node_location: Loc,
        function_name: str,
        inputs: dict[PortID, OutputLoc],
        output_list: list[PortID],
    ) -> Path:
        call_args_path = self._worker_call_args_path(node_location)
        node_definition = WorkerCallArgs(
            function_name=function_name,
            inputs={
                k: self._output_path(loc, port).relative_to(self.tkr_dir)
                for k, (loc, port) in inputs.items()
            },
            outputs={
                k: self._output_path(node_location, k).relative_to(self.tkr_dir)
                for k in output_list
            },
            output_dir=self._outputs_dir(node_location).relative_to(self.tkr_dir),
            done_path=self._done_path(node_location).relative_to(self.tkr_dir),
            error_path=self._error_path(node_location).relative_to(self.tkr_dir),
            logs_path=self.logs_path.relative_to(self.tkr_dir),
        )
        with open(call_args_path, "w+") as fh:
            fh.write(node_definition.model_dump_json())

        if (parent := node_location.parent()) is not None:
            self._metadata_path(parent).touch()

        return call_args_path.relative_to(self.tkr_dir)

    def read_worker_call_args(self, node_location: Loc) -> WorkerCallArgs:
        node_definition_path = self._worker_call_args_path(node_location)
        with open(node_definition_path, "r") as fh:
            return WorkerCallArgs(**json.load(fh))

    def link_outputs(
        self,
        new_location: Loc,
        new_port: PortID,
        old_location: Loc,
        old_port: PortID,
    ) -> None:
        new_dir = self._output_path(new_location, new_port)
        new_dir.parent.mkdir(parents=True, exist_ok=True)
        try:
            os.link(self._output_path(old_location, old_port), new_dir)
        except FileNotFoundError as e:
            logger.warning(
                f"Could not link {e.filename} to {e.filename2}."
                " Possibly a mislabelled variable?"
            )
        except OSError as e:
            raise TierkreisError(
                "Workflow already exists. Try running with a different ID or do_cleanup."
            ) from e

    def write_output(
        self, node_location: Loc, output_name: PortID, value: bytes
    ) -> Path:
        output_path = self._output_path(node_location, output_name)
        with open(output_path, "wb+") as fh:
            fh.write(bytes(value))
        return output_path

    def read_output(self, node_location: Loc, output_name: PortID) -> bytes:
        with open(self._output_path(node_location, output_name), "rb") as fh:
            return fh.read()

    def read_errors(self, node_location: Loc) -> str:
        if not self._error_logs_path(node_location).exists():
            if self._error_path(node_location).exists():
                with open(self._error_path(node_location), "r") as fh:
                    return fh.read()
            return ""
        with open(self._error_logs_path(node_location), "r") as fh:
            errors = fh.read()
        if errors == "":
            if self._error_path(node_location).exists():
                with open(self._error_path(node_location), "r") as fh:
                    return fh.read()
        return errors

    def write_node_errors(self, node_location: Loc, error_logs: str) -> None:
        with open(self._error_logs_path(node_location), "w+") as fh:
            fh.write(error_logs)

    def read_output_ports(self, node_location: Loc) -> list[PortID]:
        dir_list = list(self._outputs_dir(node_location).iterdir())
        dir_list.sort()
        return [x.name for x in dir_list if x.is_file()]

    def is_node_started(self, node_location: Loc) -> bool:
        return Path(self._nodedef_path(node_location)).exists()

    def is_node_finished(self, node_location: Loc) -> bool:
        return self._done_path(node_location).exists()

    def node_has_error(self, node_location: Loc) -> bool:
        return self._error_path(node_location).exists()

    def mark_node_finished(self, node_location: Loc) -> None:
        self._done_path(node_location).touch()

        if (parent := node_location.parent()) is not None:
            self._metadata_path(parent).touch()

    def write_metadata(self, node_location: Loc) -> None:
        with open(self._metadata_path(node_location), "w+") as fh:
            fh.write(
                json.dumps(
                    {"name": self.name, "start_time": datetime.now().isoformat()}
                )
            )

    def read_metadata(self, node_location: Loc) -> dict[str, Any]:
        with open(self._metadata_path(node_location)) as fh:
            return json.load(fh)

    def read_started_time(self, node_location: Loc) -> str | None:
        node_def = Path(self._nodedef_path(node_location))
        if not node_def.exists():
            return None
        since_epoch = node_def.stat().st_mtime
        return datetime.fromtimestamp(since_epoch).isoformat()

    def read_finished_time(self, node_location: Loc) -> str | None:
        done = Path(self._done_path(node_location))
        if not done.exists():
            return None
        since_epoch = done.stat().st_mtime
        return datetime.fromtimestamp(since_epoch).isoformat()
